from __future__ import annotations

import os
from copy import deepcopy
from typing import Any

from lxml.etree import _Element

from spinta.components import Context
from lxml import etree, objectify
from urllib.request import urlopen

from spinta.core.ufuncs import Expr
from spinta.utils.naming import to_property_name, to_model_name, Deduplicator, to_dataset_name


# mapping of XSD datatypes to DSA datatypes
# XSD datatypes: https://www.w3.org/TR/xmlschema11-2/#built-in-datatypes
# DSA datatypes: https://atviriduomenys.readthedocs.io/dsa/duomenu-tipai.html#duomenu-tipai
# TODO: finish mapping and make sure all things are mapped correctly
DATATYPES_MAPPING = {
    "string": "string",
    "boolean": "boolean",
    "decimal": "number",
    "float": "number",
    "double": "number",

    # duration has to be mapped to integer. In addition to this, we need a prepare function.
    # This prepare function takes in given duration and turns it into integer (timedelta).
    # More about time entities here:
    # https://atviriduomenys.readthedocs.io/dsa/vienetai.html#laiko-vienetai
    # TODO: add prepare function for duration
    #  https://github.com/atviriduomenys/spinta/issues/594
    "duration": "string",

    "dateTime": "datetime",
    "time": "time",
    "date": "date",
    # format: target type, target column, value in the column
    "gYearMonth": "date;enum;M",
    "gYear": "date;enum;Y",
    "gMonthDay": "string",
    "gDay": "string",
    "gMonth": "string",
    "hexBinary": "string",
    "base64Binary": "binary;prepare;base64",
    "anyURI": "uri",
    "QName": "string",
    "NOTATION": "string",
    "normalizedString": "string",
    "token": "string",
    "language": "string",
    "NMTOKEN": "string",
    "NMTOKENS": "string",
    "Name": "string",
    "NCName": "string",
    "ID": "string",
    "IDREF": "string",
    "IDREFS": "string",
    "ENTITY": "string",
    "ENTITIES": "string",
    "integer": "integer",
    "nonPositiveInteger": "integer",
    "negativeInteger": "integer",
    "long": "integer",
    "int": "integer",
    "short": "integer",
    "byte": "integer",
    "nonNegativeInteger": "integer",
    "unsignedLong": "integer",
    "unsignedInt": "integer",
    "unsignedShort": "integer",
    "unsignedByte": "integer",
    "positiveInteger": "integer",
    "yearMonthDuration": "integer",
    "dayTimeDuration": "integer",
    "dateTimeStamp": "datetime",
    "": "string",

}


class XSDModel:
    deduplicate: Deduplicator
    xsd: 'XSDReader'
    dataset_name: str
    node: _Element
    type: str = "model"
    name: str | None = None
    basename: str | None = None
    external: dict | None = None
    properties: dict | None = None
    uri: str | None = None
    description: str | None = None
    root_properties: dict | None = None
    parent_model: XSDModel | None = None
    is_root_model: bool | None = None

    """
    Class for creating and handling DSA models from XSD files.

    Example of model data:
    {
        "type": "model",
        "name": "",
        "description": "Įvairūs duomenys",
        "properties": {},
        "external": resource_model_external_info,
        "uri": "http://www.w3.org/2000/01/rdf-schema#Resource"
    }
    """

    def __init__(self, xsd: 'XSDReader', node: _Element = None):
        self.deduplicate = Deduplicator()
        self.xsd: 'XSDReader' = xsd
        self.dataset_name: str = xsd.dataset_name
        self.node: _Element = node

    def __eq__(self, other: XSDModel) -> bool:
        if self.properties == other.properties:
            return True
        return False

    def get_data(self):
        model_data: dict = {
            "type": "model",
            "name": self.name
        }
        if self.description is not None:
            model_data["description"] = self.description
        if self.properties is not None:
            model_data["properties"] = self.properties
        if self.external is not None:
            model_data["external"] = self.external
        if self.uri is not None:
            model_data["uri"] = self.uri

        return model_data

    def add_external_info(self, external_name: str):
        self.external = {
            "dataset": self.dataset_name,
            "resource": "resource1",
            "name": external_name
        }

    def set_name(self, name: str):
        self.basename = name
        self.name = f"{self.dataset_name}/{name}"

    def _get_enums(self, node: _Element) -> dict[str, dict[str, Any]]:
        enums = {}
        simple_type = node.xpath(f'./*[local-name() = "simpleType"]')
        if simple_type:
            enums = self.xsd.get_enums_from_simple_type(simple_type[0])
        else:
            node_type = node.get("type")
            if node_type and ":" in node_type:
                node_type = node_type.split(":")[1]
            if node_type in self.xsd.custom_types:
                enums = self.xsd.custom_types[node_type]["enums"]

        return enums

    def _node_to_partial_property(self, node: _Element) -> tuple[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:
        """Node can be either element or attribute.
        This function only processes things common to attributes and elements"""
        prop = dict()

        prop["description"] = XSDReader.get_description(node)
        property_name = node.get("name")
        prop["external"] = {"name": property_name}
        property_id = to_property_name(property_name)
        prop["type"] = self.xsd.get_property_type(node)
        if ";" in prop["type"]:
            prop_type, target, value = prop["type"].split(";")
            prop["type"] = prop_type
            if target == "enum":
                prop[target] = value
            if target == "prepare":
                prop["external"][target] = Expr(value)

        prop["enums"] = self._get_enums(node)

        return self.deduplicate(property_id), prop

    def attributes_to_properties(
        self,
        element: _Element
    ) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:
        properties = {}
        attributes = element.xpath(f'./*[local-name() = "attribute"]')
        complex_type = element.xpath(f'./*[local-name() = "complexType"]')
        if complex_type:
            properties.update(self.attributes_to_properties(complex_type[0]))
        for attribute in attributes:

            property_id, prop = self._node_to_partial_property(attribute)
            if not prop["type"]:
                prop["type"] = "string"
            # property source
            prop["external"]["name"] = f'@{prop["external"]["name"]}'

            # property required or not. For attributes only.
            use = attribute.get("use")
            if use == "required":
                required = True
            else:
                required = False
            prop["required"] = required
            properties[property_id] = prop

        # TODO: attribute can be a ref to an externally defined attribute also. Not in RC though
        #  https://github.com/atviriduomenys/spinta/issues/605
        return properties

    def simple_element_to_property(
        self,
        element: _Element,
        is_array: bool = False,
        source_path: str = None
    ) -> tuple[str, dict[str, str | bool | dict[str, Any]]]:
        """
        simple element is an element which is either
        an inline or simple type element and doesn't have a ref

        Example of a property:

            'properties': {
                'id': {
                    'type': 'integer',
                    'type_args': None,
                    'required': True,
                    'unique': True,
                    'external': {
                        'name': 'ID',
                        'prepare': NA,
                    }
                },
             },
        """

        property_id, prop = self._node_to_partial_property(element)
        if XSDReader.node_is_ref(element):
            ref: str = element.get("ref")
            if ":" in ref:
                ref = ref.split(":")[1]
            prop["external"]["name"] = ref
            property_id = self.deduplicate(to_property_name(ref))
        prop["external"]["name"] = f'{prop["external"]["name"]}/text()'
        if source_path:
            prop["external"]["name"] = f'{source_path}/{prop["external"]["name"]}'
        if prop.get("type") == "":
            prop["type"] = "string"
        if XSDReader.is_array(element) or is_array:
            property_id += "[]"
        if XSDReader.is_required(element):
            prop["required"] = True
        else:
            prop["required"] = False
        return property_id, prop

    def properties_from_simple_elements(
        self,
        node: _Element,
        from_root: bool = False,
        properties_required: bool = None
    ) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:
        is_array = False
        if self.xsd.is_array(node):
            is_array = True

        properties = {}
        elements = node.xpath(f'./*[local-name() = "element"]')
        for element in elements:
            if (self.xsd.node_is_simple_type_or_inline(element) and
                    not XSDReader.node_is_ref(element) and
                    not (self.xsd._node_is_referenced(element) and from_root)):
                property_id, prop = self.simple_element_to_property(element)
                if from_root:
                    prop["required"] = False
                if properties_required is True:
                    prop["required"] = True
                if properties_required is False:
                    prop["required"] = False
                if is_array:
                    property_id = f"{property_id}[]"
                properties[property_id] = prop
        return properties

    def get_text_property(self, property_type=None) -> dict[str, dict[str, str | dict[str, str]]]:
        if property_type is None:
            property_type = "string"
        return {
            self.deduplicate('text'): {
                'type': property_type,
                'external': {
                    'name': 'text()'
                }
            }}

    def has_non_ref_properties(self) -> bool:
        return any([prop["type"] not in ("ref", "backerf") for prop in self.properties.values()])

    def add_ref_property(self, ref_model):
        property_id = self.deduplicate(to_property_name(ref_model.basename))
        prop = {"type": "ref", "model": ref_model.name}
        self.properties.update({property_id: prop})


class XSDReader:

    def __init__(self, path, dataset_name: str):
        self._path: str = path
        self.models: dict[str, XSDModel] = {}
        self.custom_types: dict = {}
        self._dataset_given_name: str = dataset_name
        self._set_dataset_and_resource_info()
        self.deduplicate: Deduplicator = Deduplicator()

    def get_property_type(self, node: _Element) -> str:
        if node.get("ref"):
            return "ref"
        property_type: str = node.get("type")
        if not property_type:
            # this is a self defined simple type, so we take it's base as type
            restrictions: list = node.xpath(f'./*[local-name() = "simpleType"]/*[local-name() = "restriction"]')
            if restrictions:
                property_type = restrictions[0].get("base", "")
            else:
                property_type = ""
        # getting rid of the prefix
        if ":" in property_type:
            property_type = property_type.split(":")[1]

        if property_type in self.custom_types:
            property_type = self.custom_types.get(property_type).get("base", "")
        if property_type in DATATYPES_MAPPING:
            property_type = DATATYPES_MAPPING[property_type]
        else:
            property_type = "string"

        return property_type

    @staticmethod
    def get_enums_from_simple_type(node: _Element) -> dict[str, dict[str, Any]]:
        enums = {}
        enum_value = {}
        restrictions = node.xpath(f'./*[local-name() = "restriction"]')
        if restrictions:
            # can be enum
            enumerations = restrictions[0].xpath('./*[local-name() = "enumeration"]')
            for enumeration in enumerations:
                enum_item = {
                    "source": enumeration.get("value")
                }
                enum_value.update({enumeration.get("value"): enum_item})
            enums[""] = enum_value

        return enums

    def _extract_custom_simple_types(self, node: _Element):
        """
        format of custom types:
        {
            "type_name": {
                "base": "type_base"
            }
        }
        """
        custom_types_nodes = node.xpath(f'./*[local-name() = "simpleType"]')
        custom_types = {}
        for type_node in custom_types_nodes:
            type_name = type_node.get("name")
            restrictions = type_node.xpath(f'./*[local-name() = "restriction"]')
            property_type_base = restrictions[0].get("base", "")
            property_type_base = property_type_base.split(":")[1]

            # enums
            enums = self.get_enums_from_simple_type(type_node)

            custom_types[type_name] = {
                "base": property_type_base,
                "enums": enums
            }
        self.custom_types = custom_types

    @staticmethod
    def get_description(element: _Element) -> str:
        annotation = element.xpath(f'./*[local-name() = "annotation"]')
        if not annotation:
            annotation = element.xpath(f'./*[local-name() = "simpleType"]/*[local-name() = "annotation"]')
        description = ""
        if annotation:
            documentation = annotation[0].xpath(f'./*[local-name() = "documentation"]')
            for documentation_part in documentation:
                if documentation_part.text is not None:
                    description = f"{description}{documentation_part.text} "
        return description.strip()

    def _node_is_referenced(self, node):
        # if this node is referenced by some other node
        node_name = node.get('name')
        xpath_search_string = f'//*[@ref="{node_name}"]'
        references = self.root.xpath(xpath_search_string)

        # also check with namespace prefixes.
        # Though, it is possible that this isn't correct XSD behaviour, but it seems common in RC
        if not references:
            for prefix in self.namespaces:
                prefixed_node_name = f"{prefix}:{node_name}"
                xpath_search_string = f'//*[@ref="{prefixed_node_name}"]'
                references = self.root.xpath(xpath_search_string)
                if references:
                    return True
        if references:
            return True
        return False

    def _get_referenced_node(self, node):
        ref = node.get("ref")
        if ":" in ref:
            ref = ref.split(":")[1]
        xpath_query = f"/*/*[@name='{ref}']"
        referenced_node = self.root.xpath(xpath_query)[0]
        return referenced_node

    @staticmethod
    def node_is_ref(node: _Element) -> bool:
        if node.get("ref"):
            return True
        return False

    @staticmethod
    def is_array(element: _Element) -> bool:
        max_occurs: str = element.get("maxOccurs", "1")
        return max_occurs == "unbounded" or int(max_occurs) > 1
    
    def _extract_root(self):
        if self._path.startswith("http"):
            document = etree.parse(urlopen(self._path))
            objectify.deannotate(document, cleanup_namespaces=True)
            self.root = document.getroot()
        else:
            with open(self._path) as file:
                text = file.read()
                self.root = etree.fromstring(bytes(text, encoding='utf-8'))

    def _set_dataset_and_resource_info(self):
        self.dataset_name = to_dataset_name(os.path.splitext(os.path.basename(self._path))[0])
        self.dataset_and_resource_info = {
            'type': 'dataset',
            'name': self.dataset_name,
            'resources': {
                "resource1": {
                    'type': 'xml',
                },
            },
            'given_name': self._dataset_given_name
        }

    def _get_separate_complex_type_node_by_type(self, node_type: str) -> _Element:
        if node_type is not None:
            node_type = node_type.split(":")
            if len(node_type) > 1:
                node_type = node_type[1]
            else:
                node_type = node_type[0]
        if node_type not in DATATYPES_MAPPING:
            complex_types = self.root.xpath(f'./*[local-name() = "complexType"]')
            for node in complex_types:
                if node.get("name") == node_type:
                    return node

    def _get_separate_complex_type_node(self, node: _Element) -> _Element:
        node_type: str | list = node.get('type')
        return self._get_separate_complex_type_node_by_type(node_type)

    def _node_has_separate_complex_type(self, node: _Element) -> bool:
        node_type: str | list = node.get('type')
        if node_type is not None:
            node_type = node_type.split(":")
            if len(node_type) > 1:
                node_type = node_type[1]
            else:
                node_type = node_type[0]
            if node_type not in DATATYPES_MAPPING:
                complex_types = self.root.xpath(f'./*[local-name() = "complexType"]')
                for node in complex_types:
                    if node.get("name") == node_type:
                        return True
        return False

    def node_is_simple_type_or_inline(self, node: _Element) -> bool:
        if self._node_has_separate_complex_type(node):
            return False
        return bool(
            (node.xpath(f'./*[local-name() = "annotation"]') and len(node.getchildren()) == 1) or
            (node.xpath(f'./*[local-name() = "simpleType"]')) or
            (len(node.getchildren()) == 0)
        )

    @staticmethod
    def _is_element(node: _Element) -> bool:
        if node.xpath('local-name()') == "element":
            return True
        return False

    @staticmethod
    def is_required(element: _Element) -> bool:
        min_occurs = int(element.get("minOccurs", 1))
        if min_occurs > 0:
            nillable = element.get("nillable", False)
            if nillable == "true":
                return False
            return True
        return False

    def _build_ref_properties(self, complex_type: _Element, is_array: bool, model: XSDModel, new_source_path: str, node: _Element, referenced_element: _Element) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:
        """
        Helper method for methods properties_from_references and properties_from_type_references
        """

        properties = {}

        sequences = complex_type.xpath("./*[local-name() = 'sequence']")
        if not sequences:
            choices = complex_type.xpath("./*[local-name() = 'choice']")
            if choices and XSDReader.is_array(choices[0]):
                if len(choices[0]) == 1 and not complex_type.get("mixed") == "true":
                    is_array = True

        # NOTE: it's not fully clear, if it's a ref or a backref if `choice` `maxOccurs="unbounded" always, or only
        #  when there's only one element inside
        # if we only have one ref element and if it's inside a choice/sequence (this node) which is maxOccurs = unbounded then it's array
        # if XSDReader.is_array(node) and len(node) == 1:
        if XSDReader.is_array(node):
            is_array = True
        if XSDReader.is_array(referenced_element):
            is_array = True
        if is_array:
            property_type = "backref"
        else:
            property_type = "ref"
        referenced_model_names = self._create_model(
            referenced_element,
            source_path=new_source_path,
            parent_model=model,
        )
        for referenced_model_name in referenced_model_names:
            property_id, prop = model.simple_element_to_property(referenced_element, is_array=is_array)
            prop["external"]["name"] = prop["external"]["name"].replace("/text()", '')
            if is_array:
                if not property_id.endswith("[]"):
                    property_id += "[]"
                property_type = "backref"
            prop["type"] = property_type
            prop["model"] = f"{referenced_model_name}"
            properties[property_id] = prop

        return properties

    def _properties_from_type_references(
        self,
        node: _Element,
        model: XSDModel,
        source_path: str = ""
    ) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:

        properties = {}

        for typed_element in node.xpath('./*[local-name() = "element"]'):

            new_source_path = source_path

            if typed_element.xpath(f'./*[local-name() = "complexType"]') \
                    or self._node_has_separate_complex_type(typed_element):

                if typed_element.xpath(f'./*[local-name() = "complexType"]'):
                    complex_type = typed_element.xpath(f'./*[local-name() = "complexType"]')[0]
                if self._node_has_separate_complex_type(typed_element):
                    complex_type = self._get_separate_complex_type_node(typed_element)

                is_array = False

                referenced_element = typed_element

                # avoiding recursion
                if referenced_element.get("name") in source_path.split("/"):
                    continue

                built_properties = self._build_ref_properties(complex_type, is_array, model, new_source_path, node,
                                                              referenced_element)
                properties.update(built_properties)

        return properties

    def _properties_from_references(
        self,
        node: _Element,
        model: XSDModel,
        source_path: str = ""
    ) -> dict[str, dict[str, str | bool | dict[str, str | dict[str, Any]]]]:

        properties = {}

        for ref_element in node.xpath("./*[@ref]"):
            new_source_path = source_path
            referenced_element = self._get_referenced_node(ref_element)

            # avoiding recursion
            if referenced_element.get("name") in source_path.split("/"):
                continue

            if self.node_is_simple_type_or_inline(referenced_element):
                is_array = XSDReader.is_array(ref_element)
                property_id, prop = model.simple_element_to_property(referenced_element, is_array=is_array)
                if not XSDReader.is_required(ref_element):
                    prop["required"] = False
                properties[property_id] = prop
            else:
                is_array = False

                if XSDReader.is_array(ref_element):
                    is_array = True

                if XSDReader.is_array(node):
                    is_array = True

                complex_type = referenced_element.xpath("./*[local-name() = 'complexType']")[0]

                built_properties = self._build_ref_properties(complex_type, is_array, model, new_source_path, node, referenced_element)
                for prop in built_properties.values():
                    if not XSDReader.is_required(ref_element):
                        prop["required"] = False
                properties.update(built_properties)

        return properties

    def _split_choice(
        self,
        node: _Element,
        source_path: str,
        parent_model: XSDModel,
        is_root_model: bool = False
    ) -> list[str]:
        """
        If there are choices in the element,
        we need to split it and create a separate model per each choice
        """

        model_names = []
        node_copy = deepcopy(node)
        if self._node_has_separate_complex_type(node_copy):
            complex_type_node = self._get_separate_complex_type_node(node_copy)
        else:
            complex_type_node = node_copy.xpath(f'./*[local-name() = "complexType"]')[0]

        choice_nodes = complex_type_node.xpath(f'./*[local-name() = "choice"]')

        # if maxOccurs=unbound and there's a second `choice` inside, we have to split that one
        if choice_nodes and choice_nodes[0].get("maxOccurs") == "unbounded":
            choice_nodes = choice_nodes[0].xpath(f'./*[local-name() = "choice"]')
        if choice_nodes:
            choice_node = choice_nodes[0]
        else:
            choice_node = complex_type_node.xpath(f'./*[local-name() = "sequence"]/*[local-name() = "choice"]')[0]
        if len(choice_node) > 0:
            choice_node_copy = deepcopy(choice_node)
            choice_node_parent = choice_node.getparent()
            choice_node_parent.remove(choice_node)

            for choice in choice_node_copy:
                if complex_type_node.xpath(f'./*[local-name() = "sequence"]') and choice_node_copy.xpath(
                        f'./*[local-name() = "sequence"]'):
                    choice_copy = deepcopy(choice)
                    for node_in_choice in choice:
                        choice_node_parent.insert(0, node_in_choice)
                    returned_model_names = self._create_model(
                        node_copy,
                        source_path=source_path,
                        parent_model=parent_model,
                        is_root_model=is_root_model
                    )

                    model_names.extend(returned_model_names)

                    for node_in_choice in choice_copy:
                        node_in_choice = choice_node_parent.xpath(f"./*[@name=\'{node_in_choice.get('name')}\']")[0]
                        choice_node_parent.remove(node_in_choice)
                else:
                    choice_node_parent.insert(0, choice)
                    returned_model_names = self._create_model(
                        node_copy,
                        source_path=source_path,
                        parent_model=parent_model,
                        is_root_model=is_root_model
                    )
                    model_names.extend(returned_model_names)

                    choice_node_parent.remove(choice)
        return model_names

    def _create_model(
        self,
        node: _Element,
        source_path: str = "",
        is_root_model: bool = False,
        parent_model: XSDModel = None,
    ) -> list[str]:
        """
        Parses an element and makes a model out of it. If it is a complete model, it will be added to the models list.
        """
        model = XSDModel(self)
        model.parent_model = parent_model

        # properties of this model
        properties = {}
        properties.update(model.attributes_to_properties(node))

        new_source_path = f"{source_path}/{node.get('name')}"

        model.set_name(self.deduplicate(to_model_name(node.get("name"))))

        if node.xpath(f'./*[local-name() = "complexType"]') or self._node_has_separate_complex_type(node):

            if self._node_has_separate_complex_type(node):
                complex_type_node = self._get_separate_complex_type_node(node)
            else:
                complex_type_node = node.xpath(f'./*[local-name() = "complexType"]')[0]

            # if there is choices, we need to create a separate model for each choice
            choices = complex_type_node.xpath(f'./*[local-name() = "choice"]')
            # if choices is unbounded, we treat it like sequence
            if not choices or choices[0].get("maxOccurs") == "unbounded":
                if choices:
                    choices = complex_type_node.xpath(f'./*[local-name() = "choice"]/*[local-name() = "choice"]')
                else:
                    choices = complex_type_node.xpath(f'./*[local-name() = "sequence"]/*[local-name() = "choice"]')
            if choices and choices[0].get("maxOccurs") != "unbounded":
                return self._split_choice(
                    node,
                    source_path=source_path,
                    parent_model=parent_model,
                    is_root_model=is_root_model
                )

            # if complextype node's property mixed is true, it allows text inside
            if complex_type_node.get("mixed") == "true":
                properties.update(model.get_text_property())

            # if this is complexType node which has complexContent, with a separate
            # node, we need to join the contents of them both
            if complex_type_node.xpath(f'./*[local-name() = "complexContent"]'):
                complex_type_node = complex_type_node.xpath(f'./*[local-name() = "complexContent"]/*[local-name() = "extension"]')[0]
                complex_content_base_name = complex_type_node.get("base")
                complex_content_base_node = self._get_separate_complex_type_node_by_type(complex_content_base_name)
                if complex_content_base_node.xpath(f'./*[local-name() = "sequence"]'):
                    sequence_node = complex_content_base_node.xpath(f'./*[local-name() = "sequence"]')[0]
                    properties.update(model.properties_from_simple_elements(sequence_node))

            if (
                complex_type_node.xpath(f'./*[local-name() = "sequence"]') or
                complex_type_node.xpath(f'./*[local-name() = "all"]') or
                complex_type_node.xpath(f'./*[local-name() = "simpleContent"]') or
                choices or
                len(complex_type_node) > 0
            ):
                """
                source: https://stackoverflow.com/questions/36286056/the-difference-between-all-sequence-choice-and-group-in-xsd
                    When to use xsd:all, xsd:sequence, xsd:choice, or xsd:group:

                    Use xsd:all when all child elements must be present, independent of order.
                    Use xsd:sequence when child elements must be present per their occurrence constraints and order does matters.
                    Use xsd:choice when one of the child element must be present.
                    Use xsd:group as a way to wrap any of the above in order to name and reuse in multiple locations within an XSD.
                    Note that occurrence constraints can appear on xsd:all, xsd:sequence, or xsd:choice in addition to the 
                    child elements to achieve various cardinality effects. For example, if minOccurs="0" were added to xsd:element children of xsd:all, element order would be insignificant,
                    but not all child elements would have to be present.
                """
                properties_required = None
                if complex_type_node.xpath(f'./*[local-name() = "sequence"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "sequence"]')
                elif complex_type_node.xpath(f'./*[local-name() = "all"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "all"]')
                elif complex_type_node.xpath(f'./*[local-name() = "choice"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "choice"]')
                    properties_required = False
                elif complex_type_node.xpath(f'./*[local-name() = "simpleContent"]'):
                    sequence_or_all_nodes = complex_type_node.xpath(f'./*[local-name() = "simpleContent"]/*[local-name() = "extension"]')

                    text_property_type = sequence_or_all_nodes[0].get("base")
                    if text_property_type is not None:
                        if ":" in text_property_type:
                            text_property_type = text_property_type.split(":")[1]
                        text_property_type = DATATYPES_MAPPING[text_property_type]
                    text_property = model.get_text_property(text_property_type)
                    properties.update(text_property)
                    # additionally extract properties from attributes, if it's let's say, simpleContent node
                    properties.update(model.attributes_to_properties(sequence_or_all_nodes[0]))
                else:
                    sequence_or_all_nodes = None
                if sequence_or_all_nodes:
                    sequence_or_all_node = sequence_or_all_nodes[0]
                else:
                    sequence_or_all_node = complex_type_node

                properties.update(model.properties_from_simple_elements(
                    sequence_or_all_node,
                    properties_required=properties_required))

                # references
                properties_from_references = self._properties_from_references(
                    sequence_or_all_node,
                    model=model,
                    source_path=new_source_path)
                if properties_required is False:
                    for prop in properties_from_references.values():
                        prop["required"] = False
                properties.update(properties_from_references)

                # complex type child nodes - to models
                properties_from_references = self._properties_from_type_references(
                    sequence_or_all_node,
                    model=model,
                    source_path=new_source_path)
                if properties_required is False:
                    for prop in properties_from_references.values():
                        prop["required"] = False
                properties.update(properties_from_references)

        model.properties = properties

        if properties:

            model.add_external_info(external_name=new_source_path)
            model.description = self.get_description(node)
            self.models[model.name] = model

            return [model.name, ]

        return []

    def _add_resource_model(self):
        resource_model = XSDModel(self)
        resource_model.add_external_info(external_name="/")
        resource_model.type = "model"
        resource_model.description = "Įvairūs duomenys"
        resource_model.uri = "http://www.w3.org/2000/01/rdf-schema#Resource"
        resource_model.properties = resource_model.properties_from_simple_elements(self.root, from_root=True)
        resource_model.root_properties = {}
        if resource_model.properties:
            resource_model.set_name(self.deduplicate(f"Resource"))
            self.models[resource_model.name] = resource_model

    def _parse_root_node(self):
        for node in self.root:
            if (
                    self._is_element(node) and
                    (not self.node_is_simple_type_or_inline(node) or self.node_is_ref(node)) and
                    not self._node_is_referenced(node)
            ):
                self._create_model(node, is_root_model=True)

    def _extract_namespaces(self):
        self.namespaces = self.root.nsmap

    def _add_refs_for_backrefs(self):
        for model in self.models.values():
            for property_id, prop in model.properties.items():
                if prop["type"] == "backref":
                    referenced_model = self.models[prop["model"]]
                    # checking if the ref already exists.
                    # They can exist multiple times, but refs should be added only once
                    prop_added = False
                    for prop in referenced_model.properties.values():
                        if "model" in prop and prop["model"] == model.name:
                            prop_added = True
                    if not prop_added:
                        referenced_model.add_ref_property(model)

    def _sort_properties_alphabetically(self):
        for model in self.models.values():
            model.properties = dict(sorted(model.properties.items()))

    def _remove_sources_from_secondary_models(self):
        """
        Only models which represent root elements for XML need to have source
        """

        for parsed_model in self.models.values():

            # we need to add root properties to properties if it's a root model
            if parsed_model.parent_model is not None and parsed_model.parent_model.name in self.models:
                parsed_model.external["name"] = ""

    def _remove_duplicate_models(self):
        """removes models that are exactly the same"""

        do_loop = True

        do_not_remove = []

        while do_loop:

            model_pairs = {}

            for model_name, model in self.models.items():
                for another_model_name, another_model in self.models.items():
                    if model is not another_model and model == another_model and another_model_name not in do_not_remove:
                        if (
                            another_model_name not in model_pairs.values() and
                            another_model_name not in model_pairs
                        ):
                            model_pairs[another_model_name] = model_name

            for another_model_name, model_name in model_pairs.items():
                parent_model = self.models[another_model_name].parent_model
                if parent_model and parent_model.name in self.models:
                    for property_id, prop in parent_model.properties.items():
                        if "model" in prop and prop["model"] == another_model_name:
                            prop["model"] = model_name
                    self.models.pop(another_model_name)
                else:
                    do_not_remove.append(another_model_name)
                do_not_remove.append(model_name)

            if not model_pairs:
                do_loop = False
            else:
                print(self.dataset_name)
                for old_model, new_model in model_pairs.items():
                    print(f"{old_model.split('/')[1]} -> {new_model.split('/')[1]}")

    def start(self):
        self._extract_root()

        # preparation part

        self._extract_namespaces()
        self._extract_custom_simple_types(self.root)

        # main part

        self._add_resource_model()

        self._parse_root_node()

        # models transformations

        self._remove_unneeded_models()

        self.models = dict(sorted(self.models.items()))

        self._remove_duplicate_models()

        self._compile_nested_properties()

        self._remove_sources_from_secondary_models()

        self._add_refs_for_backrefs()

        self._sort_properties_alphabetically()

    def remove_extra_root_models(self, model: XSDModel) -> XSDModel:
        """
        removes root models that have only one property from the root
        """
        stop_removing = False

        while not stop_removing:
            # remove the model itself if it's a root proxy model
            if (len(model.properties) == 1) and (list(model.properties.values())[0]["type"] in ("ref", "backref")):
                model = self.models[list(model.properties.values())[0]["model"]]
                model.parent_model = None
            else:
                stop_removing = True

        return model

    def _remove_proxy_models(self, model: XSDModel):
        """ Removes models which have only one property
            Usually these are proxy models to indicate arrays, but there can be other situations
            Removes the models that are in the middle of other models
            or at the end and have one property, then this property is joined to the referring model.
        """

        self.new_models[model.name] = model

        new_properties = {}
        for property_id, prop in model.properties.items():
            if prop["type"] in ("ref", "backref"):
                referee = self.models[prop["model"]]
                parse_referee = True
                while len(referee.properties) == 1:
                    ref_property_id, ref_prop = list(referee.properties.items())[0]

                    # if it's not a ref, this means that it's a final property, and we add it as a property itself
                    if ref_prop["type"] not in ("ref", "backref"):
                        prop["external"]["name"] = f'{prop["external"]["name"]}/{ref_prop["external"]["name"]}'

                        # also transfer all attributes of the property
                        prop["required"] = ref_prop["required"]
                        if ref_prop["description"]:
                            prop["description"] = ref_prop["description"]

                        is_array = False
                        if prop["type"] == "backref":
                            is_array = True

                        prop["type"] = ref_prop["type"]
                        del prop["model"]
                        property_id = ref_property_id

                        if is_array:
                            property_id = f"{property_id}[]"

                        parse_referee = False
                        break

                    if prop["type"] == "backref" and ref_prop["type"] == "backref":
                        # basically, do nothing
                        break
                    else:
                        referee = self.models[ref_prop["model"]]
                        referee.parent_model = model
                        if prop["type"] == "ref" and ref_prop["type"] == "backref":
                            prop["type"] = "backref"
                            property_id = f"{property_id}[]"
                        if "external" in prop and "external" in ref_prop:
                            prop["external"]["name"] = f'{prop["external"]["name"]}/{ref_prop["external"]["name"]}'
                            prop["model"] = ref_prop["model"]


                if not self._has_backref(model, referee) and parse_referee:
                    self._remove_proxy_models(referee)

            new_properties[property_id] = prop
        model.properties = new_properties

    def _remove_unneeded_models(self):
        """
        Proxy models are those that have only one property which is a ref to another model.
        They can act as placeholders, or as array indicators.
        If either one of them is an array, drop the proxy model and replace the reference to point to the new model
        If both referencing properties are not arrays, the resulting model shouldn't be an array, and if any of them is an array, the resulting ref is an array (backref).
        If both models, the referrer and the referee are arrays, do not drop them, because this means that it's an array of arrays.
        """
        self.new_models = {}
        for model_name, model in self.models.items():

            # we need to start from root models
            if model.parent_model is None:
                model = self.remove_extra_root_models(model)
                self._remove_proxy_models(model)

        self.models = self.new_models

    def _has_backref(self, model: XSDModel, ref_model: XSDModel) -> bool:
        has_backref = False
        for ref_model_property in ref_model.properties.values():
            if (ref_model_property.get("type") == "backref") and (ref_model_property.get('model') == model.name):
                has_backref = True
        return has_backref

    def _add_model_nested_properties(self, root_model: XSDModel, model: XSDModel, property_prefix: str = "", source_path: str = ""):
        """recursively gather nested properties or root model"""
        # go forward, add property prefix, which is constructed rom properties that came rom beore models, and construct pathh orward also
        # probably will need to cut beginning for path sometimes

        source_path = source_path.lstrip("/")

        root_properties = {}

        properties = deepcopy(model.properties)

        for property_id, prop in properties.items():

            if (model != root_model and
                    not ("model" in prop and prop["model"] == model.parent_model.name)):

                # update property source and name and add it to the root properties
                if property_prefix:
                    property_id = f"{property_prefix}.{property_id}"

                if "external" in prop and source_path:
                    prop["external"]["name"] = f"{source_path}/{prop['external']['name']}"

                root_properties[property_id] = prop

            # if this property is ref or backref, gather the properties of the model to which it points
            # (if it's not the root model or back pointing ref)
            if "model" in prop:
                ref_model = self.models[prop['model']]

                # there are two types of ref - direct, and for backref. We don't want to traverse the ones or backref,
                # because it will create an infinite loop.
                # If it's a ref, we need to build the path from the root of the model. If it's backref (array) -
                # it's relative to array
                if prop["type"] == "ref":
                    if self._has_backref(model, ref_model):
                        continue
                    if source_path:
                        new_source_path = f"{source_path}/{ref_model.external['name'].replace(model.external['name'], '').lstrip('/')}"
                    else:
                        new_source_path = ref_model.external['name'].replace(model.external['name'], '')
                    new_source_path = new_source_path.lstrip('/')

                # property type is backref
                else:
                    new_source_path = ""

                self._add_model_nested_properties(root_model, ref_model, property_prefix=property_id, source_path=new_source_path)

        root_model.properties.update(root_properties)

    def _compile_nested_properties(self):
        for parsed_model in self.models.values():

            # we need to add root properties to properties if it's a root model
            if parsed_model.parent_model is None or parsed_model.parent_model.name not in self.models:

                self._add_model_nested_properties(parsed_model, parsed_model)


def read_schema(
    context: Context,
    path: str,
    prepare: str = None,
    dataset_name: str = ''
) -> dict[Any, dict[str, str | dict[str, str | bool | dict[str, str | dict[str, Any]]]]]:
    """
    This reads XSD schema from the url provided in path and yields asd schema models

    For now this is adjusted for XSD schemas of Registrų centras

    Elements can be:
    1. Simple type inline
    2. Simple type
    3. Complex type
    4. Complex type described separately

    Elements can have annotations, which we add as description to either a model or a property.

    1. Simple type.
        a) not a reference. It's a property
        b) a reference
            i. maxOccurs = 1
                A. both referencing and referenced elements has annotations. A ref property
                B. Only one or none of the elements has annotations. A part of the path
            ii. maxOccurs > 1 - a ref property
    2. Inline type.

        a) inline type or a custom type referencing to simple type. A property
        b) complex type defined separately of this element.

        If it's a root element, it's then added as a property to a special Resource model.

        Simple type can have annotation too.

    3. Complex type
        If element is complex type, it can be either a property or a model,
        or a part of a source path for a property or model, depending on other factors

            a) element has attributes. Then it's a model and attributes are properties.
            b) element has a sequence
                i. There is more than one element in the sequence. Then it's a model/end part of the model.
                   We treat child elements in the same way as we do root elements (recursion) and build a path
                ii. There is only one element in the sequence. It can then be a property, a model or an intermediate
                    in the path of a model.
                    Options:
                        A. That element isn't a ref.
                            We treat child elements in the same way as we do root elements (recursion) and build a path
                        B. That element is a ref.
                            B1. If maxOccurs > 1, this is a model, and that other element is also a model
                            B2. If maxOccurs = 1
                                B21. If both have annotation, then it's a model
                                B22. If only one of them has annotation, it's a part of a path


            c) element has a choice.

    4. complex type described separately

    We will build a list of models
    (another option was to make a generator for parsing models, and going deeper, but that would
    be more complex when returning models. although this option is also possible, but it can
    be reworked into this from an option with a list)

    If some model has references in the sequence, we need to also add those as models,
    mark their type as backref. In this case, we add them at the moment we meet them
    as xsd refs, because this way we will know that they are backrefs. We also need to add refs on
    those models to the current model.

    Element can be as a ref in more than one other element.

    Other things to handle: Resource model, custom types, enumerations, choices

    Element can be turned into:
        1. A property (including reference properties)
        2. A model (including referenced models)
        3. A part of another model (as a path)
        4. A part of another property (as a path)

    Attribute can only be turned into a property

    A property can also be text()

    -----------Nested properties-------------

    Root model or models can have nested properties if they have any properties that point to other models.

    """
    xsd = XSDReader(path, dataset_name)

    xsd.start()

    yield None, xsd.dataset_and_resource_info

    for model_name, parsed_model in xsd.models.items():

        yield None, parsed_model.get_data()



from spinta import commands
from spinta.components import Context
from spinta.manifests.internal_sql.components import InternalSQLManifest
import sqlalchemy as sa

from spinta.manifests.internal_sql.helpers import get_table_structure


@commands.bootstrap.register(Context, InternalSQLManifest)
def bootstrap(context: Context, manifest: InternalSQLManifest):
    store = context.get('store')
    url = sa.engine.make_url(manifest.path)
    url.get_dialect()
    engine = sa.create_engine(url)
    inspector = sa.inspect(engine)
    meta = sa.MetaData(engine)
    if not inspector.has_table('_manifest'):
        table = get_table_structure(meta)
        table.create()

    for backend in store.backends.values():
        commands.bootstrap(context, backend)



from copy import copy
from typing import Any, List, Union, Generator
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Iterator
from typing import Tuple
from typing import TypeVar

from spinta import commands
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.store import load_manifest
from spinta.components import Context, Property, Node, Namespace
from spinta.components import Mode
from spinta.core.config import RawConfig, Path
from spinta.core.config import ResourceTuple
from spinta.core.config import parse_resource_args
from spinta.core.context import configure_context, create_context
from spinta.datasets.components import Dataset, Resource, Attribute, Entity, ExternalBackend
from spinta.dimensions.prefix.components import UriPrefix
from spinta.exceptions import InvalidResourceSource
from spinta.manifests.components import Manifest
from spinta.manifests.components import ManifestPath
from spinta.manifests.helpers import get_manifest_from_type, init_manifest
from spinta.types.datatype import Ref, DataType, Array, Object, Denorm
from spinta.utils.naming import Deduplicator
from spinta.utils.schema import NA
from spinta.utils.schema import NotAvailable
from spinta.components import Model


def create_manifest_from_inspect(
    context: Context = None,
    manifest: Union[str, ManifestPath] = None,
    resources: Tuple = None,
    backend: str = None,
    formula: str = None,
    dataset: str = None,
    auth: str = None,
    priority: str = "manifest",
    only_url: bool = False,
):
    # Reset Context, to not have any outside influence from given
    if context is None:
        rc = RawConfig()
        rc.read([Path('spinta', 'spinta.config:CONFIG')])
        context = create_context(name="inspect", rc=rc)

    has_manifest_priority = priority == 'manifest'
    if resources:
        resources = parse_resource_args(*resources, formula)

    context = configure_context(
        context,
        [manifest] if manifest else None,
        mode=Mode.external,
        backend=backend
    )
    with context:
        require_auth(context, auth)
        store = load_manifest(context, ensure_config_dir=True, full_load=True)
        old = store.manifest
        manifest = Manifest()
        init_manifest(context, manifest, 'inspect')
        commands.merge(context, manifest, manifest, old, has_manifest_priority)

        if not resources:
            resources = []
            for ds in commands.get_datasets(context, old).values():
                for resource in ds.resources.values():
                    external = resource.external
                    if external == '' and resource.backend:
                        external = resource.backend.config['dsn']
                    if not any(res.external == external for res in resources):
                        if only_url and not ("http://" in external or "https://" in external):
                            raise InvalidResourceSource(source=external)
                        resources.append(ResourceTuple(type=resource.type, external=external, prepare=resource.prepare))

        if resources:
            for resource in resources:
                _merge(context, manifest, manifest, resource, has_manifest_priority, dataset)

        # Sort models for render
        sorted_models = {}
        for key, model in commands.get_models(context, manifest).items():
            if key not in sorted_models.keys():
                if model.external and model.external.resource:
                    resource = model.external.resource
                    for resource_key, resource_model in resource.models.items():
                        if resource_key not in sorted_models.keys():
                            sorted_models[resource_key] = resource_model
                else:
                    sorted_models[key] = model
        commands.set_models(context, manifest, sorted_models)
    return context, manifest


def _merge(context: Context, manifest: Manifest, old: Manifest, resource: ResourceTuple, has_manifest_priority: bool, dataset: str = None):
    rc: RawConfig = context.get('rc')
    manifest_ = get_manifest_from_type(rc, resource.type)
    path = ManifestPath(type=manifest_.type, path=resource.external, prepare=resource.prepare)
    context = configure_context(context, [path], mode=Mode.external, dataset=dataset)
    store = load_manifest(context, full_load=True)
    new = store.manifest
    commands.merge(context, manifest, old, new, has_manifest_priority)


@commands.merge.register(Context, Manifest, Manifest, Manifest, bool)
def merge(context: Context, manifest: Manifest, old: Manifest, new: Manifest, has_manifest_priority: bool) -> None:
    backends = zipitems(
        old.backends.values(),
        new.backends.values(),
        _backend_dsn
    )

    backend_deduplicator = Deduplicator("_{}")
    for backend in backends:
        for o, n in backend:
            if not (o and n):
                if o:
                    backend_deduplicator(o.name)
                if n:
                    name = backend_deduplicator(n.name)
                    n.name = name
            merge(context, manifest, o, n)
    datasets = zipitems(
        commands.get_datasets(context, old).values(),
        commands.get_datasets(context, new).values(),
        _dataset_key,
    )
    deduplicator = Deduplicator("{}")
    for ds in datasets:
        for o, n in ds:
            if o:
                deduplicator(o.name)
            elif n and not o:
                name = deduplicator(n.name)
                n.name = name
                n.ns.name = name
            commands.merge(context, manifest, o, n, has_manifest_priority)


@commands.merge.register(Context, Manifest, NotAvailable, ExternalBackend)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: ExternalBackend) -> None:
    manifest.backends[new.name] = new


@commands.merge.register(Context, Manifest, ExternalBackend, ExternalBackend)
def merge(context: Context, manifest: Manifest, old: ExternalBackend, new: ExternalBackend) -> None:
    manifest.backends[old.name] = old


@commands.merge.register(Context, Manifest, ExternalBackend, NotAvailable)
def merge(context: Context, manifest: Manifest, old: ExternalBackend, new: NotAvailable) -> None:
    manifest.backends[old.name] = old


@commands.merge.register(Context, Manifest, NotAvailable, Dataset, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Dataset, has_manifest_priority: bool) -> None:
    commands.set_dataset(context, manifest, new.name, new)
    _merge_resources(context, manifest, old, new)

    dataset_models = _filter_models_for_dataset(context, new.manifest, new)
    deduplicator = Deduplicator()
    for model in dataset_models:
        model.name = deduplicator(model.name)
        merge(context, manifest, NA, model, has_manifest_priority)

    new.manifest = manifest


@commands.merge.register(Context, Manifest, Dataset, Dataset, bool)
def merge(context: Context, manifest: Manifest, old: Dataset, new: Dataset, has_manifest_priority: bool) -> None:
    old.id = coalesce(old.id, new.id)
    old.description = coalesce(old.description, new.description)
    old.lang = coalesce(old.lang, new.lang)
    old.name = coalesce(old.name, new.name)
    old.manifest = coalesce(old.manifest, new.manifest)
    old.website = coalesce(old.website, new.website)
    old.source = coalesce(old.source, new.source)
    old.given = coalesce(old.given, new.given)
    old.level = coalesce(old.level, new.level)
    old.access = coalesce(old.access, new.access)
    old.title = coalesce(old.title, new.title)
    old.resources = coalesce(old.resources, new.resources)
    old.prefixes = coalesce(old.prefixes, new.prefixes)
    if old.ns and new.ns and old.ns.name == new.ns.name:
        commands.merge(context, manifest, old.ns, new.ns)
    else:
        old.ns = coalesce(old.ns, new.ns)
    commands.set_dataset(context, manifest, old.name, old)
    _merge_prefixes(context, manifest, old, new)
    _merge_resources(context, manifest, old, new)

    dataset_models = _filter_models_for_dataset(context, manifest, old)
    models = zipitems(
        dataset_models,
        commands.get_models(context, new.manifest).values(),
        _model_key
    )
    resource_list = []
    for res in new.resources.values():
        resource_list.append(_resource_key(res))
    deduplicator = Deduplicator()
    for model in models:
        for om, nm in model:
            if om:
                deduplicator(om.name)
            if om and not nm:
                if om.external and om.external.resource:
                    if _resource_key(om.external.resource) in resource_list:
                        om.external.name = None
            if not om and nm:
                if nm.external:
                    nm.external.dataset = old
                    name = deduplicator(f"{nm.external.dataset.name}/{nm.basename}")
                    nm.name = name
                    nm.ns.name = nm.external.dataset.name
            merge(context, manifest, om, nm, has_manifest_priority)


@commands.merge.register(Context, Manifest, Dataset, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Dataset, new: NotAvailable, has_manifest_priority: bool) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, UriPrefix)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: UriPrefix) -> None:
    dataset = new.parent
    commands.get_dataset(context, manifest, dataset.name).prefixes[new.name] = new


@commands.merge.register(Context, Manifest, UriPrefix, UriPrefix)
def merge(context: Context, manifest: Manifest, old: UriPrefix, new: UriPrefix) -> None:
    old.type = coalesce(new.type, old.type)
    old.id = coalesce(new.id, old.id)
    old.uri = coalesce(new.uri, old.uri)
    old.title = coalesce(new.title, old.title)
    old.description = coalesce(new.description, old.description)

    old.name = coalesce(old.name, new.name)


@commands.merge.register(Context, Manifest, UriPrefix, NotAvailable)
def merge(context: Context, manifest: Manifest, old: UriPrefix, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Namespace)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Namespace) -> None:
    commands.set_namespace(context, manifest, new.name, new)


@commands.merge.register(Context, Manifest, Namespace, Namespace)
def merge(context: Context, manifest: Manifest, old: Namespace, new: Namespace) -> None:
    old.keymap = coalesce(new.keymap, old.keymap)
    old.backend = coalesce(new.backend, old.backend)
    old.parent = coalesce(new.parent, old.parent)
    old.description = coalesce(new.description, old.description)
    old.generated = coalesce(new.generated, old.generated)
    old.lang = coalesce(new.lang, old.lang)
    old.enums = coalesce(new.enums, old.enums)

    old.name = coalesce(old.name, new.name)
    old.given = coalesce(old.given, new.given)
    old.title = coalesce(old.title, new.title)
    old.models = coalesce(old.models, new.models)
    old.access = coalesce(old.access, new.access)


@commands.merge.register(Context, Manifest, Namespace, NotAvailable)
def merge(context: Context, manifest: Manifest, old: Namespace, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Resource)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Resource) -> None:
    commands.get_dataset(context, manifest, new.dataset.name).resources[new.name] = new


@commands.merge.register(Context, Manifest, Resource, Resource)
def merge(context: Context, manifest: Manifest, old: Resource, new: Resource) -> None:
    old.type = coalesce(new.type, old.type)
    old.description = coalesce(new.description, old.description)
    old.comments = coalesce(new.comments, old.comments)
    old.lang = coalesce(new.lang, old.lang)

    old.backend = coalesce(old.backend, new.backend)
    old.dataset = coalesce(old.dataset, new.dataset)
    old.models = old.models
    old.eid = coalesce(old.eid, new.eid)
    old.given = coalesce(old.given, new.given)
    old.access = coalesce(old.access, new.access)
    old.title = coalesce(old.title, new.title)
    old.name = coalesce(old.name, new.name)
    old.level = coalesce(old.level, new.level)
    old.external = old.external
    old.prepare = coalesce(old.prepare, new.prepare)


@commands.merge.register(Context, Manifest, Resource, NotAvailable)
def merge(context: Context, manifest: Manifest, old: Resource, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Model, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Model, has_manifest_priority: bool) -> None:
    old = copy(new)
    old.external = copy(old.external)
    old_name = old.name
    if f'{old.ns.name}/{old.basename}' != old.name and old.ns.name:
        name = f'{old.ns.name}/{old.basename}'
        old.name = name
        new.name = name
    if old.external and old.external.resource:
        resources = zipitems(
            old.external.dataset.resources.values(),
            [old.external.resource],
            _resource_key
        )
        for res in resources:
            for old_res, new_res in res:
                if old_res and new_res:
                    old.external.resource = old_res
                    if old_name != old.name:
                        if old_name in old_res.models:
                            del old_res.models[old_name]
                    old_res.models[old.name] = old
    old.manifest = manifest
    commands.set_model(context, manifest, old.name, old)
    _merge_model_properties(context, manifest, old, new, has_manifest_priority)


@commands.merge.register(Context, Manifest, Model, Model, bool)
def merge(context: Context, manifest: Manifest, old: Model, new: Model, has_manifest_priority: bool) -> None:
    old.description = coalesce(new.description, old.description)
    old.comments = coalesce(new.comments, old.comments)
    old.lang = coalesce(new.lang, old.lang)

    old.id = coalesce(old.id, new.id)
    old.eid = coalesce(old.eid, new.eid)
    old.base = coalesce(old.base, new.base)
    old.given = coalesce(old.given, new.given)
    old.name = coalesce(old.name, new.name)
    old.title = coalesce(old.title, new.title)
    old.access = coalesce(old.access, new.access)
    old.level = coalesce(old.level, new.level)
    old.external = coalesce(old.external, new.external)
    old.manifest = manifest

    commands.set_model(context, manifest, old.name, old)
    _merge_model_properties(context, manifest, old, new, has_manifest_priority)

    if old.external and new.external:
        if not has_manifest_priority or (old.external.unknown_primary_key and not new.external.unknown_primary_key):
            keys = zipitems(
                old.properties.values(),
                new.external.pkeys,
                _property_key
            )
            new_keys = []
            for key in keys:
                for o, n in key:
                    if o and n:
                        new_keys.append(o)
            if new_keys:
                old.external.pkeys = new_keys
                old.external.unknown_primary_key = new.external.unknown_primary_key


@commands.merge.register(Context, Manifest, Model, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Model, new: NotAvailable, has_manifest_priority: bool) -> None:
    if old.external and not old.external.name:
        for prop in old.properties.values():
            if prop.external:
                prop.external.name = None


@commands.merge.register(Context, Manifest, NotAvailable, Property, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Property, has_manifest_priority: bool) -> None:
    if new.external:
        new.external.prop = new
    new.dtype.prop = new
    merge(context, manifest, new.dtype, new.dtype)
    new.model.properties[new.name] = new


@commands.merge.register(Context, Manifest, Property, Property, bool)
def merge(context: Context, manifest: Manifest, old: Property, new: Property, has_manifest_priority: bool) -> None:
    if not has_manifest_priority:
        merged = old
        merged.type = coalesce(new.type, old.type)
        merged.uri = coalesce(new.uri, old.uri)
        merged.description = coalesce(new.description, old.description)
        merged.enum = coalesce(new.enum, old.enum)
        merged.enums = coalesce(new.enums, old.enums)
        merged.unit = coalesce(new.unit, old.unit)
        merged.comments = coalesce(new.comments, old.comments)
        merged.lang = coalesce(new.lang, old.lang)

        merged.model = coalesce(old.model, new.model)
        merged.external = coalesce(old.external, new.external)
        merged.place = coalesce(old.place, new.place)
        merged.name = coalesce(old.name, new.name)
        merged.given = coalesce(old.given, new.given)
        merged.title = coalesce(old.title, new.title)
        merged.level = coalesce(old.level, new.level)
        merged.access = coalesce(old.access, new.access)
        merged.dtype.prop = merged

        old.model.properties[old.name] = merged

        merge(context, manifest, merged.dtype, new.dtype)


@commands.merge.register(Context, Manifest, Property, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Property, new: NotAvailable, has_manifest_priority: bool) -> None:
    if old.external:
        old.external.name = None
    model = commands.get_model(context, manifest, old.model.name)
    model.properties[old.name] = old


@commands.merge.register(Context, Manifest, DataType, Array)
def merge(context: Context, manifest: Manifest, old: DataType, new: Array) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)
    models = zipitems(
        [merged.items.model],
        commands.get_models(context, manifest).values(),
        _model_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                properties = zipitems(
                    merged.items,
                    o.properties.values(),
                    _property_key
                )
                for prop in properties:
                    for po, pn in prop:
                        if po and pn:
                            merged.items = po
                            break
                    else:
                        continue
                    break
            break
        else:
            continue
        break

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Object)
def merge(context: Context, manifest: Manifest, old: DataType, new: Object) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    new_properties = {}
    for key, value in merged.properties.items():
        new_key = key
        new_value = value
        models = zipitems(
            [value.model],
            commands.get_models(context, manifest).values(),
            _model_key
        )
        for model in models:
            for o, n in model:
                if o and n:
                    properties = zipitems(
                        merged.items,
                        o.properties.values(),
                        _property_key
                    )
                    for prop in properties:
                        for po, pn in prop:
                            if po and pn:
                                new_value = po
                                new_key = po.name
                                break
                        else:
                            continue
                        break
                break
            else:
                continue
            break
        new_properties[new_key] = new_value
    merged.properties = new_properties

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Ref)
def merge(context: Context, manifest: Manifest, old: DataType, new: Ref) -> None:
    merged = copy(new)
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    models = zipitems(
        [merged.model],
        commands.get_models(context, manifest).values(),
        _model_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                if n.external and merged.prop.model.external:
                    if n.external.dataset is merged.prop.model.external.dataset:
                        merged.model = n
                        if not n.external.unknown_primary_key and n.external.pkeys:
                            merged.refprops = n.external.pkeys
                        else:
                            new_refprops = []
                            properties = zipitems(
                                merged.refprops,
                                n.properties.values(),
                                _property_key
                            )
                            for prop in properties:
                                for po, pn in prop:
                                    if po and pn:
                                        new_refprops.append(pn)

                            if not new_refprops:
                                new_refprops = merged.refprops
                            merged.refprops = new_refprops
                        break
        else:
            continue
        break
    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Denorm)
def merge(context: Context, manifest: Manifest, old: DataType, new: Denorm) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    models = zipitems(
        [merged.rel_prop.model],
        commands.get_models(context, manifest).values(),
        _model_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                properties = zipitems(
                    merged.items,
                    o.properties.values(),
                    _property_key
                )
                for prop in properties:
                    for po, pn in prop:
                        if po and pn:
                            merged.rel_prop = po
                            break
                    else:
                        continue
                    break
            break
        else:
            continue
        break

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, DataType)
def merge(context: Context, manifest: Manifest, old: DataType, new: DataType) -> None:
    merged = new
    merged.name = coalesce(new.name, old.name)
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)
    old.prop.dtype = merged


def _filter_models_for_dataset(
    context: Context,
    manifest: Manifest,
    dataset: Dataset
) -> List[Model]:
    models = []
    for model in commands.get_models(context, manifest).values():
        if model.external:
            if model.external.dataset is dataset:
                models.append(model)
    return models


def _merge_model_properties(
    context: Context,
    manifest: Manifest,
    old: Model,
    new: Model,
    has_manifest_priority: bool
):
    properties = zipitems(
        old.properties.values(),
        new.properties.values(),
        _property_key,
    )
    deduplicator = Deduplicator("_{}")
    for prop in properties:
        for o, n in prop:
            if n:
                n = copy(n)
                n.model = new
            if o:
                deduplicator(o.basename)
                o.model = old
            if n and not o:
                name = deduplicator(n.basename)
                n.model = old
                n.name = name
                n.place = name
            commands.merge(context, manifest, o, n, has_manifest_priority)


def _merge_prefixes(
    context: Context,
    manifest: Manifest,
    old: Dataset,
    new: Dataset
):
    prefixes = zipitems(
        old.prefixes.values(),
        new.prefixes.values(),
        _name_key,
    )
    for prefix in prefixes:
        for o, n in prefix:
            commands.merge(context, manifest, o, n)


def _merge_resources(
    context: Context,
    manifest: Manifest,
    old: Dataset,
    new: Dataset
):
    old_resources = [] if old == NA else old.resources.values()
    new_resources = [] if new == NA else new.resources.values()
    resources = zipitems(
        old_resources,
        new_resources,
        _resource_key,
    )
    deduplicator = Deduplicator("{}")
    for res in resources:
        for o, n in res:
            if o:
                deduplicator(o.name)
            elif n:
                n.name = deduplicator(n.name)
            commands.merge(context, manifest, o, n)


TItem = TypeVar('TItem')


class PriorityKey:
    id: Any = None
    name: str = None
    source: Any = None

    def __init__(self, _id=None, name=None, source=None):
        self.id = _id
        self.name = name
        self.source = source

    def __eq__(self, other):
        if isinstance(other, PriorityKey):
            if self.id and other.id:
                if self.id == other.id:
                    return True
            if self.name and other.name:
                if self.name == other.name:
                    return True
            if self.source and other.source:
                if isinstance(other.source, tuple) or isinstance(self.source, tuple):
                    if set(self.source).issubset(other.source) or set(other.source).issubset(self.source):
                        return True
                else:
                    if self.source == other.source:
                        return True
        return False

    def __str__(self):
        return f"PriorityKey(id: {self.id}, name: {self.name}, source: {self.source})"

    def __hash__(self):
        return hash(str(self))


def zipitems(
    a: TItem,
    b: TItem,
    key: Callable[[TItem], Hashable],
) -> Iterator[List[Tuple[TItem, TItem]]]:

    res: Dict[
        Hashable,  # key
        List[
            List[
                Any,   # a
                Any,   # b
            ]
        ]
        ,
    ] = {}

    # Map first values
    for value in a:
        value_key = key(value)
        new_value = [value, NA]

        if isinstance(value_key, Tuple):
            for existing_key in res.keys():
                if set(value_key) == set(existing_key):
                    value_key = existing_key
                    break
        elif isinstance(value_key, PriorityKey):
            for existing_key in res.keys():
                if value_key == existing_key:
                    value_key = existing_key
                    break

        res.setdefault(value_key, []).append(new_value)

    # Map second value
    for value in b:
        value_key = key(value)
        new_value = [NA, value]

        mapped_keys = []
        if isinstance(value_key, Tuple):
            value_key_set = set(value_key)
            mapped_keys = [existing_key for existing_key in res if value_key_set.issubset(set(existing_key))]
        elif isinstance(value_key, PriorityKey):
            # Only get the first match
            for existing_key in res:
                if existing_key == value_key:
                    mapped_keys = [existing_key]
                    break

        if not mapped_keys:
            mapped_keys = [value_key]

        for mapped_key in mapped_keys:
            if mapped_key not in res or not res[mapped_key]:
                res[mapped_key] = [new_value]
                continue

            extension_list = []
            for existing_value in res[mapped_key]:
                if existing_value[1] is NA:
                    existing_value[1] = value
                else:
                    extension_list.append([existing_value[0], value])
            res[mapped_key] += extension_list

    yield from res.values()


def coalesce(*args: Any) -> Any:
    for arg in args:
        if arg:
            return arg
    return arg


def _name_key(node: Node) -> str:
    return node.name


def _dataset_key(dataset: Dataset) -> PriorityKey:
    key = PriorityKey()
    if dataset.id:
        key.id = dataset.id
    if dataset.given.name:
        key.name = dataset.given.name
    key.source = _dataset_resource_source_key(dataset)
    return key


def _dataset_resource_source_key(dataset: Dataset) -> Tuple:
    keys = []
    for resource in dataset.resources.values():
        keys.append(_resource_source_key(resource))
    return tuple(keys)


def _property_key(prop: Property) -> PriorityKey:
    key = PriorityKey()
    if prop.id:
        key.id = prop.id
    if prop.given.name:
        key.name = prop.given.name
    key.source = _property_source_key(prop)
    return key


def _property_source_key(prop: Property) -> str:
    if prop.external and prop.external.name:
        return prop.external.name
    else:
        return prop.name


def _resource_key(resource: Resource) -> PriorityKey:
    key = PriorityKey()
    if resource.id:
        key.id = resource.id
    if resource.given.name:
        key.name = resource.given.name
    key.source = _resource_source_key(resource)
    return key


def _resource_source_key(resource: Resource) -> str:
    result = resource.name
    manifest = resource.dataset.manifest
    if resource.external:
        result = resource.external
    else:
        if manifest.backends:
            if resource.backend.name in manifest.backends:
                result = manifest.backends[resource.backend.name].config['dsn']
        elif manifest.store.backends:
            if resource.backend.name in manifest.store.backends:
                result = manifest.store.backends[resource.backend.name].config['dsn']
    if "@" in result:
        result = result.split("@")[1]
    return result


def _model_key(model: Model) -> PriorityKey:
    key = PriorityKey()
    if model.id:
        key.id = model.id
    if model.given.name:
        key.name = model.given.name
    key.source = _model_source_key(model)
    return key


def _model_source_key(model: Model) -> str:
    if model.external and model.external.name:
        return f'{_resource_source_key(model.external.resource)}/{model.external.name}'
    else:
        return model.name


def _backend_dsn(backend: ExternalBackend) -> str:
    dsn = backend.config['dsn']
    if "@" in dsn:
        dsn = dsn.split("@")[1]
    return dsn






import geoalchemy2 as ga
import sqlalchemy as sa

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name, get_pg_name
from spinta.components import Context
from spinta.types.geometry.components import Geometry


@commands.prepare.register(Context, PostgreSQL, Geometry)
def prepare(
    context: Context,
    backend: PostgreSQL,
    dtype: Geometry,
    **kwargs
) -> list:
    prop = dtype.prop
    name = get_column_name(prop)

    # Have to disable spatial index and create our own, because Geoalchemy2 uses their own naming convention
    kwargs = {
        'geometry_type': dtype.geometry_type,
        'srid': dtype.srid,
        'spatial_index': False
    }
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    column_type = ga.Geometry(**kwargs)
    nullable = not dtype.required
    columns = [
        column := sa.Column(name, column_type, nullable=nullable),
        sa.Index(None, column, postgresql_using='GIST')
    ]

    # Technically speaking we shouldn't create UniqueConstraint and Index at the same time
    # behind the scenes UniqueConstraint is an Index
    # but this uses GIST for Index, not BTREE
    # need to investigate this further
    if dtype.unique:
        columns.append(
            sa.UniqueConstraint(column)
        )
    return columns



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import sqlalchemy as sa
from sqlalchemy.sql.functions import Function

from spinta.components import Model
from spinta.core.ufuncs import Expr
from spinta.datasets.backends.sql.components import Sql
from spinta.exceptions import UnknownMethod
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, QueryPage, QueryParams
from spinta.ufuncs.basequerybuilder.helpers import merge_with_page_selected_list, \
    merge_with_page_sort, merge_with_page_limit
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.itertools import ensure_list


class SqlFrom:
    backend: Sql
    joins: Dict[str, sa.Table]
    from_: sa.Table

    def __init__(self, backend: Sql, table: sa.Table):
        self.backend = backend
        self.joins = {}
        self.from_ = table

    def get_table(
        self,
        env: SqlQueryBuilder,
        prop: ForeignProperty,
    ) -> sa.Table:
        fpr: Optional[ForeignProperty] = None
        for fpr in prop.chain:
            if fpr.name in self.joins:
                continue

            # Left table foreign keys
            lmodel = fpr.left.prop.model
            if len(fpr.chain) > 1:
                # Use table alias of previous join.
                ltable = self.joins[fpr.chain[-2].name]
            else:
                # Use the main table, without alias.
                ltable = self.backend.get_table(lmodel)
            lenv = env(model=lmodel, table=ltable)
            lfkeys = lenv.call('join_table_on', fpr.left.prop)
            lfkeys = ensure_list(lfkeys)

            # Right table primary keys
            rpkeys = []
            rmodel = fpr.right.prop.model
            rtable = self.backend.get_table(rmodel).alias()
            renv = env(model=rmodel, table=rtable)
            for rpk in fpr.left.refprops:
                rpkeys += ensure_list(
                    renv.call('join_table_on', rpk)
                )

            # Number of keys on both left and right must be equal.
            assert len(lfkeys) == len(rpkeys), (lfkeys, rpkeys)
            condition = []
            for lfk, rpk in zip(lfkeys, rpkeys):
                condition += [lfk == rpk]

            # Build `JOIN rtable ON (condition)`.
            assert len(condition) > 0
            if len(condition) == 1:
                condition = condition[0]
            else:
                condition = sa.and_(*condition)

            self.joins[fpr.name] = rtable
            self.from_ = self.from_.outerjoin(rtable, condition)

        return self.joins[fpr.name]


class SqlQueryBuilder(BaseQueryBuilder):
    backend: Sql
    model: Model
    table: sa.Table
    joins: SqlFrom
    columns: List[sa.Column]

    def init(self, backend: Sql, table: sa.Table, params: QueryParams = None):
        result = self(
            backend=backend,
            table=table,
            columns=[],
            resolved={},
            selected=None,
            joins=SqlFrom(backend, table),
            sort=[],
            limit=None,
            offset=None,
            distinct=False,
            group_by=None,
            page=QueryPage(),
        )
        result.init_query_params(params)
        return result

    def build(self, where):
        if self.selected is None:
            # If select list was not explicitly given by client, then select all
            # properties.
            self.call('select', Expr('select'))
        merged_selected = merge_with_page_selected_list(self.columns, self.page)
        merged_sorted = merge_with_page_sort(self.sort, self.page)
        merged_limit = merge_with_page_limit(self.limit, self.page)
        qry = sa.select(merged_selected)
        qry = qry.select_from(self.joins.from_)
        if where is not None:
            qry = qry.where(where)

        if merged_sorted:
            qry = qry.order_by(*merged_sorted)

        if merged_limit is not None:
            qry = qry.limit(merged_limit)

        if self.offset is not None:
            qry = qry.offset(self.offset)

        if self.group_by is not None:
            qry = qry.group_by(*self.group_by)

        if self.distinct:
            qry = qry.distinct()
        return qry

    def execute(self, expr: Any):
        expr = self.call('_resolve_unresolved', expr)
        return super().execute(expr)

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(name=expr.name, expr=str(expr(*args, **kwargs)))

    def add_column(self, column: Union[sa.Column, Function]) -> int:
        """Returns position in select column list, which is stored in
        Selected.item.
        """
        assert isinstance(column, (sa.Column, Function)), column
        if column not in self.columns:
            self.columns.append(column)
        return self.columns.index(column)



from __future__ import annotations

import itertools
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union
from typing import cast
from typing import overload

from spinta import commands
from spinta import exceptions
from spinta.auth import authorized
from spinta.backends.constants import BackendFeatures
from spinta.backends.nobackend.components import NoBackend
from spinta.commands import authorize
from spinta.commands import check
from spinta.commands import load
from spinta.components import Action, Component, PageBy, Page, PageInfo, UrlParams, pagination_enabled
from spinta.components import Base
from spinta.components import Context
from spinta.components import Mode
from spinta.components import Model
from spinta.components import Property
from spinta.core.access import link_access_param
from spinta.core.access import load_access_param
from spinta.core.enums import Level
from spinta.datasets.components import ExternalBackend
from spinta.dimensions.comments.helpers import load_comments
from spinta.dimensions.enum.components import EnumValue
from spinta.dimensions.enum.components import Enums
from spinta.dimensions.enum.helpers import link_enums
from spinta.dimensions.enum.helpers import load_enums
from spinta.dimensions.lang.helpers import load_lang_data
from spinta.dimensions.param.helpers import load_params
from spinta.exceptions import KeymapNotSet, InvalidLevel
from spinta.exceptions import PropertyNotFound
from spinta.exceptions import UndefinedEnum
from spinta.exceptions import UnknownPropertyType
from spinta.hacks.urlparams import extract_params_sort_values
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.components import PropertyRow
from spinta.nodes import get_node
from spinta.nodes import load_model_properties
from spinta.nodes import load_node
from spinta.types.helpers import check_model_name
from spinta.types.helpers import check_property_name
from spinta.types.namespace import load_namespace_from_name
from spinta.ufuncs.loadbuilder.components import LoadBuilder
from spinta.ufuncs.loadbuilder.helpers import page_contains_unsupported_keys, get_allowed_page_property_types
from spinta.units.helpers import is_unit
from spinta.utils.enums import enum_by_value
from spinta.utils.schema import NA

if TYPE_CHECKING:
    from spinta.datasets.components import Attribute


def _load_namespace_from_model(context: Context, manifest: Manifest, model: Model):
    ns = load_namespace_from_name(context, manifest, model.name)
    ns.models[model.model_type()] = model
    model.ns = ns


@load.register(Context, Model, dict, Manifest)
def load(
    context: Context,
    model: Model,
    data: dict,
    manifest: Manifest,
    *,
    source: Manifest = None,
) -> Model:
    model.parent = manifest
    model.manifest = manifest
    model.mode = manifest.mode  # TODO: mode should be inherited from namespace.
    load_node(context, model, data)
    model.lang = load_lang_data(context, model.lang)
    model.comments = load_comments(model, model.comments)
    model.given.name = data.get("given_name", None)
    if model.keymap:
        model.keymap = manifest.store.keymaps[model.keymap]
    else:
        model.keymap = manifest.keymap

    _load_namespace_from_model(context, manifest, model)
    load_access_param(model, data.get('access'), itertools.chain(
        [model.ns],
        model.ns.parents(),
    ))
    load_level(model, model.level)
    load_model_properties(context, model, Property, data.get('properties'))

    # XXX: Maybe it is worth to leave possibility to override _id access?
    model.properties['_id'].access = model.access

    config = context.get('config')

    if model.base:
        base: dict = model.base
        model.base = get_node(
            context,
            config,
            manifest,
            model.eid,
            base,
            group='nodes',
            ctype='base',
            parent=model,
        )
        load_node(context, model.base, base)
        model.base.model = model
        commands.load(context, model.base, base, manifest)
    if model.unique:
        unique_properties = []
        for unique_set in model.unique:
            prop_set = []
            for prop_name in unique_set:
                if "." in prop_name:
                    prop_name = prop_name.split(".")[0]
                if prop_name not in model.properties:
                    raise PropertyNotFound(model, property=prop_name)
                prop_set.append(model.properties[prop_name])
            if prop_set:
                unique_properties.append(prop_set)
        model.unique = unique_properties

    if model.external:
        external: dict = model.external
        model.external = get_node(
            context,
            config,
            manifest,
            model.eid,
            external,
            group='datasets',
            ctype='entity',
            parent=model,
        )
        model.external.model = model
        load_node(context, model.external, external, parent=model)
        commands.load(context, model.external, external, manifest)
        model.given.pkeys = external.get('pk', [])
    else:
        model.external = None
        model.given.pkeys = []

    builder = LoadBuilder(context)
    builder.update(model=model)
    builder.load_page()

    model.params = load_params(context, manifest, model.params)

    if not model.name.startswith('_') and not model.basename[0].isupper():
        raise Exception(model.basename, "MODEL NAME NEEDS TO BE UPPER CASED")

    return model


@load.register(Context, Base, dict, Manifest)
def load(context: Context, base: Base, data: dict, manifest: Manifest) -> None:
    load_level(base, data['level'])


@overload
@commands.link.register(Context, Model)
def link(context: Context, model: Model):
    # Link external source.
    if model.external:
        commands.link(context, model.external)

        if model.keymap is None:
            raise KeymapNotSet(model)

    # Link model backend.
    if model.backend and isinstance(model.backend, str):
        if model.backend in model.manifest.backends:
            model.backend = model.manifest.backends[model.backend]
        else:
            model.backend = model.manifest.store.backends[model.backend]
    elif (
        model.mode == Mode.external and
        model.external and
        model.external.resource and
        model.external.resource.backend
    ):
        model.backend = model.external.resource.backend
    elif model.mode == Mode.external:
        model.backend = NoBackend()
    else:
        model.backend = model.manifest.backend

    if model.external and model.external.dataset:
        link_access_param(model, itertools.chain(
            [model.external.dataset],
            [model.ns],
            model.ns.parents(),
        ))
    else:
        link_access_param(model, itertools.chain(
            [model.ns],
            model.ns.parents(),
        ))

    # Link base
    if model.base:
        commands.link(context, model.base)

    # Link model properties.
    for prop in model.properties.values():
        commands.link(context, prop)

    _link_model_page(model)


def _disable_page_in_model(model: Model):
    model.page.enabled = False


def _link_model_page(model: Model):
    if not model.backend or not model.backend.supports(BackendFeatures.PAGINATION):
        _disable_page_in_model(model)
        return
    # Disable page if external backend and model.ref not given
    if isinstance(model.backend, ExternalBackend):
        if (model.external and not model.external.name) or not model.external:
            _disable_page_in_model(model)
            return

        if '_id' in model.page.keys:
            model.page.keys.pop('_id')
        if '-_id' in model.page.keys:
            model.page.keys.pop('-_id')
    else:
        # Force '_id' to be page key if other keys failed the checks
        if not model.page.enabled and page_contains_unsupported_keys(model.page):
            model.page.keys = {'_id': model.properties['_id']}
            model.page.enabled = True

        # Add _id to internal, if it's not added
        if '_id' not in model.page.keys and '-_id' not in model.page.keys:
            model.page.keys['_id'] = model.properties["_id"]

    if len(model.page.keys) == 0:
        _disable_page_in_model(model)


@overload
@commands.link.register(Context, Base)
def link(context: Context, base: Base):
    base.parent = commands.get_model(context, base.model.manifest, base.parent)
    base.pk = [
        base.parent.properties[pk]
        for pk in base.pk
    ] if base.pk else []
    if commands.identifiable(base):
        if base.pk and base.pk != base.parent.external.pkeys:
            base.parent.add_keymap_property_combination(base.pk)


@load.register(Context, Property, dict, Manifest)
def load(
    context: Context,
    prop: Property,
    data: PropertyRow,
    manifest: Manifest,
) -> Property:
    config = context.get('config')
    prop.type = 'property'
    prop, data = load_node(context, prop, data, mixed=True)
    prop = cast(Property, prop)

    parents = list(itertools.chain(
        [prop.model, prop.model.ns],
        prop.model.ns.parents(),
    ))
    load_access_param(prop, prop.access, parents)
    prop.enums = load_enums(context, [prop] + parents, prop.enums)
    prop.lang = load_lang_data(context, prop.lang)
    prop.comments = load_comments(prop, prop.comments)
    if prop.prepare_given:
        prop.given.prepare = prop.prepare_given
    load_level(prop, prop.level)

    if data['type'] is None:
        raise UnknownPropertyType(prop, type=data['type'])
    if data['type'] == 'ref' and not commands.identifiable(prop):
        data['type'] = '_external_ref'
    prop.dtype = get_node(
        context,
        config,
        manifest,
        prop.model.eid,
        data,
        group='types',
        parent=prop,
    )
    prop.dtype.type = 'type'
    prop.dtype.prop = prop
    load_node(context, prop.dtype, data)
    if prop.model.external:
        prop.external = _load_property_external(context, manifest, prop, prop.external)
    else:
        prop.external = NA
    commands.load(context, prop.dtype, data, manifest)
    # if prop.model.unique:
    #     if isinstance(prop.dtype, Ref):
    #         if '.id' not in prop.name:
    #             prop.model.unique = [list(map(lambda val: val.replace(
    #                 prop.name, prop.name + '._id'), val)) for val in prop.model.unique]
    unit: Optional[str] = prop.enum
    if unit is None:
        prop.given.enum = None
        prop.given.unit = None
        prop.enum = None
        prop.unit = None
    elif is_unit(prop.dtype, unit):
        prop.given.enum = None
        prop.given.unit = unit
        prop.enum = None
        prop.unit = unit
    else:
        prop.given.enum = unit
    prop.given.explicit = prop.explicitly_given if prop.explicitly_given is not None else True
    prop.given.name = prop.given_name
    return prop


def load_level(
    component: Component,
    given_level: Union[Level, int, str],
):
    if given_level:
        if isinstance(given_level, Level):
            level = given_level
        else:
            if isinstance(given_level, str) and given_level.isdigit():
                given_level = int(given_level)
            if not isinstance(given_level, int):
                raise InvalidLevel(component, level=given_level)
            level = enum_by_value(component, 'level', Level, given_level)
    else:
        level = None
    component.level = level


def _link_prop_enum(
    prop: Property,
) -> Optional[EnumValue]:
    if prop.given.enum:
        enums: List[Enums] = (
            [prop.enums] +
            [prop.model.ns.enums] +
            [ns.enums for ns in prop.model.ns.parents()] +
            [prop.model.manifest.enums]
        )
        for enums_ in enums:
            if enums_ and prop.given.enum in enums_:
                return enums_[prop.given.enum]
        if not is_unit(prop.dtype, prop.given.enum):
            raise UndefinedEnum(prop, name=prop.given.enum)
    elif prop.enums:
        return prop.enums.get('')


@overload
@commands.link.register(Context, Property)
def link(context: Context, prop: Property):
    commands.link(context, prop.dtype)
    if prop.external:
        if isinstance(prop.external, list):
            for external in prop.external:
                commands.link(context, external)
        else:
            commands.link(context, prop.external)

    model = prop.model

    if prop.model.external and prop.model.external.dataset:
        parents = list(itertools.chain(
            [model],
            [model.external.dataset],
            [model.ns],
            model.ns.parents(),
        ))
    else:
        parents = list(itertools.chain(
            [model],
            [model.ns],
            model.ns.parents(),
        ))
    link_access_param(prop, parents, use_given=not prop.name.startswith('_'))
    link_enums([prop] + parents, prop.enums)
    prop.enum = _link_prop_enum(prop)


def _load_property_external(
    context: Context,
    manifest: Manifest,
    prop: Property,
    data: Any,  # external data
) -> Attribute:
    if not isinstance(data, dict):
        return _load_property_external(context, manifest, prop, {'name': data})

    config = context.get('config')
    external: Attribute = get_node(
        context,
        config,
        manifest,
        prop.model.eid,
        data,
        group='datasets',
        ctype='attribute',
        parent=prop,
    )
    load_node(context, external, data, parent=prop)
    commands.load(context, external, data, manifest)
    return external


@load.register(Context, Model, dict)
def load(context: Context, model: Model, data: dict) -> dict:
    # check that given data does not have more keys, than model's schema
    non_hidden_keys = []
    for key, prop in model.properties.items():
        if not prop.hidden:
            non_hidden_keys.append(key)

    unknown_props = set(data.keys()) - set(non_hidden_keys)
    if unknown_props:
        raise exceptions.MultipleErrors(
            exceptions.FieldNotInResource(model, property=prop)
            for prop in sorted(unknown_props)
        )

    result = {}
    for name, prop in model.properties.items():
        value = data.get(name, NA)
        value = load(context, prop.dtype, value)
        if value is not NA:
            result[name] = value
    return result


@load.register(Context, Property, object)
def load(context: Context, prop: Property, value: object) -> object:
    # This load function should be called only with /Model/prop (subresource)
    value = _prepare_prop_data(prop.name, value)
    value[prop.name] = load(context, prop.dtype, value[prop.name])
    return value


def _prepare_prop_data(name: str, data: dict):
    return {
        **{
            k: v
            for k, v in data.items()
            if k.startswith('_') and k not in ('_id', '_content_type')
        },
        name: {
            k: v
            for k, v in data.items()
            if not k.startswith('_') or k in ('_id', '_content_type')
        }
    }


@check.register(Context, Model)
def check(context: Context, model: Model):
    check_model_name(context, model)
    if '_id' not in model.properties:
        raise exceptions.MissingRequiredProperty(model, prop='_id')

    for prop in model.properties.values():
        commands.check(context, prop)


@check.register(Context, Property)
def check(context: Context, prop: Property):
    check_property_name(context, prop)
    if prop.enum:
        for value, item in prop.enum.items():
            commands.check(context, item, prop.dtype, item.prepare)

    commands.check(context, prop.dtype)


@authorize.register(Context, Action, Model)
def authorize(context: Context, action: Action, model: Model):
    authorized(context, model, action, throw=True)


@authorize.register(Context, Action, Property)
def authorize(context: Context, action: Action, prop: Property):
    authorized(context, prop, action, throw=True)


@overload
@commands.get_error_context.register(Model)
def get_error_context(model: Model, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(model.manifest, prefix=f'{prefix}.manifest')
    context['schema'] = f'{prefix}.get_eid_for_error_context()'
    context['model'] = f'{prefix}.name'
    context['dataset'] = f'{prefix}.external.dataset.name'
    context['resource'] = f'{prefix}.external.resource.name'
    context['resource.backend'] = f'{prefix}.external.resource.backend.name'
    context['entity'] = f'{prefix}.external.name'
    return context


@overload
@commands.get_error_context.register(Property)
def get_error_context(prop: Property, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(prop.model, prefix=f'{prefix}.model')
    context['property'] = f'{prefix}.place'
    context['attribute'] = f'{prefix}.external.name'
    return context


@overload
@commands.get_error_context.register(Page)
def get_error_context(prop: Page, *, prefix='this') -> Dict[str, str]:
    context = commands.get_error_context(prop.model, prefix=f'{prefix}.model')
    context['page'] = f'{prefix}.get_repr_for_error()'
    return context


@commands.identifiable.register(Model)
def identifiable(model: Model):
    return model.level is None or model.level >= Level.identifiable


@commands.identifiable.register(Base)
def identifiable(base: Base):
    return base.level is None or base.level >= Level.identifiable


@commands.identifiable.register(Property)
def identifiable(prop: Property):
    return prop.level is None or prop.level >= Level.identifiable


@commands.create_page.register(PageInfo)
def create_page(page_info: PageInfo) -> Page:
    return Page(
        model=page_info.model,
        size=page_info.size,
        enabled=page_info.enabled,
        by={key: PageBy(prop) for key, prop in page_info.keys.items()}
    )


@commands.create_page.register(PageInfo, UrlParams)
def create_page(page_info: PageInfo, params: UrlParams) -> Page:
    params_page = params.page

    enabled = pagination_enabled(page_info.model, params)
    sort_values = extract_params_sort_values(
        page_info.model,
        params
    )
    if sort_values is None:
        enabled = False

    page_by = {key: PageBy(prop) for key, prop in page_info.keys.items()}
    if enabled and sort_values:
        new_order = {}
        allowed_types = get_allowed_page_property_types()
        for key, value in sort_values.items():
            if not isinstance(value.dtype, allowed_types):
                enabled = False
                break

            new_order[key] = PageBy(value)

        for key, value in page_by.items():
            reversed_key = key[1:] if key.startswith("-") else f'-{key}'
            if key not in new_order and reversed_key not in new_order:
                new_order[key] = value

        page_by = new_order

    page = Page(
        model=page_info.model,
        size=params_page and params_page.size or page_info.size,
        enabled=enabled,
        by=page_by
    )

    if enabled and params_page and params_page.values:
        page.update_values_from_list(params_page.values)

    return page


@commands.create_page.register(PageInfo, dict)
def create_page(page_info: PageInfo, data: dict) -> Page:
    page = commands.create_page(page_info)
    model = page_info.model
    for key, value in data.items():
        prop = model.properties.get(key)
        if prop:
            page.add_prop(key, prop, value)
    return page


@commands.create_page.register(PageInfo, (list, set, tuple))
def create_page(page_info: PageInfo, data: Union[list, set, tuple]) -> Page:
    page = commands.create_page(page_info)
    page.update_values_from_list(data)
    return page


@commands.create_page.register(PageInfo, object)
def create_page(page_info: PageInfo, data: Any) -> Page:
    return commands.create_page(page_info)



import os
import tempfile
from uuid import UUID

from spinta import commands
from spinta.auth import check_scope, Scopes
from spinta.cli.helpers.store import prepare_manifest
from spinta.cli.helpers.migrate import MigrateRename, MigrateMeta
from spinta.components import Context, UrlParams, Store, Model, Config, Property
from starlette.requests import Request

from spinta.core.context import configure_context, create_context
from spinta.datasets.commands.check import check_dataset_name
from spinta.datasets.inspect.helpers import zipitems
from spinta.exceptions import NotSupportedManifestType, InvalidSchemaUrlPath, InvalidName, UnknownContentType, \
    FileSizeTooLarge, DatasetNameMissmatch, DatasetSchemaRequiresIds, \
    ModifySchemaRequiresFile, ModifyOneDatasetSchema
from spinta.manifests.components import ManifestPath, Manifest
from spinta.manifests.tabular.helpers import datasets_to_tabular
from spinta.utils.schema import NA
from spinta.utils.types import is_str_uuid
from starlette.responses import JSONResponse


def _clean_up_file(file_path: str):
    os.unlink(file_path)


def _setup_context(main_context: Context, manifest_path) -> Context:
    context = create_context("Target Manifest", rc=main_context.get('rc'))
    context = configure_context(context, [manifest_path])
    return context


def validate_manifest(context: Context, manifest: Manifest, dataset_name: str):
    datasets = commands.get_datasets(context, manifest)

    if len(datasets) != 1:
        raise ModifyOneDatasetSchema(manifest, given_amount=len(datasets))

    for dataset in datasets.values():
        if dataset.name != dataset_name:
            raise DatasetNameMissmatch(dataset, expected_dataset=dataset_name, given_dataset=dataset.name)

    rows = datasets_to_tabular(context, manifest)
    for row in rows:
        if row['id'] is not None and not is_str_uuid(row['id']):
            raise DatasetSchemaRequiresIds(manifest)


def create_migrate_rename_mapping(old_context: Context, new_context: Context, old_manifest: Manifest, new_manifest: Manifest, dataset_name: str):
    if not commands.has_dataset(old_context, old_manifest, dataset_name):
        return {}

    old_models = commands.get_dataset_models(old_context, old_manifest, dataset_name)
    new_models = commands.get_dataset_models(new_context, new_manifest, dataset_name)
    models = zipitems(
        old_models.values(),
        new_models.values(),
        _id_model_key
    )
    mapped_data = {}
    for model in models:
        for old, new in model:
            if old is not NA and new is not NA:
                possible_rename = {}
                if new.name != old.name:
                    possible_rename[''] = new.name

                filtered_old_properties = [prop for prop in old.properties.values() if not prop.name.startswith("_")]
                filtered_new_properties = [prop for prop in new.properties.values() if not prop.name.startswith("_")]
                properties = zipitems(
                    filtered_old_properties,
                    filtered_new_properties,
                    _id_property_key
                )

                for prop in properties:
                    for old_prop, new_prop in prop:
                        if old_prop is not NA and new_prop is not NA:
                            if old_prop.name != new_prop.name:
                                possible_rename[old_prop.name] = new_prop.name

                if possible_rename:
                    mapped_data[old.name] = possible_rename
    return mapped_data


def _id_model_key(model: Model):
    id_ = model.id
    return _id_key(id_)


def _id_property_key(prop: Property):
    id_ = prop.id
    return _id_key(id_)


def _id_key(id_):
    if isinstance(id_, UUID):
        id_ = str(id_)
    return id_


def _parse_and_validate_dataset_name(context: Context, manifest: Manifest, params: UrlParams):
    ns = params.model
    if isinstance(ns, Model):
        raise InvalidSchemaUrlPath(ns)

    if ns is not None:
        dataset_name = ns.name
    else:
        dataset_name = '/'.join(params.path_parts)

    if not commands.has_dataset(context, manifest, dataset_name):
        if commands.has_namespace(context, manifest, dataset_name):
            ns = commands.get_namespace(context, manifest, dataset_name)
            raise InvalidSchemaUrlPath(ns)

    if not check_dataset_name(dataset_name):
        raise InvalidName(name=dataset_name, type='dataset')

    return dataset_name


async def _create_and_validate_tmp_file(context: Context, manifest: Manifest, request: Request) -> str:
    headers = request.headers
    if 'content-type' not in headers:
        raise ModifySchemaRequiresFile()
    if 'content-type' in headers and headers['content-type'] != 'text/csv':
        raise UnknownContentType(content_type=headers['content-type'], supported_content_types=['text/csv'])

    config: Config = context.get('config')
    max_size = config.max_api_file_size
    data_stream = request.stream()

    file_path = None
    file_is_valid = True
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        file_path = tmp.name
        async for data_block in data_stream:
            tmp.write(data_block)
            if tmp.tell() > max_size * 1e+6:
                file_is_valid = False
                break

    if not file_is_valid and file_path is not None:
        _clean_up_file(file_path)
        raise FileSizeTooLarge(manifest, allowed_amount=max_size, measure='MB')

    if file_path is None:
        raise Exception("NO FILE PATH")

    return file_path


def reset_affected_objects(context: Context, manifest: Manifest, dataset_name: str):
    objects = manifest.get_objects()
    if dataset_name in objects['dataset']:
        del objects['dataset'][dataset_name]

    for key, model in objects['model'].copy().items():
        if model.external and model.external.dataset and model.external.dataset.name == dataset_name:
            del objects['model'][key]


async def schema_api(context: Context, request: Request, params: UrlParams):
    check_scope(context, Scopes.SCHEMA_WRITE)

    store: Store = context.get('store')
    manifest = store.manifest

    if not manifest.dynamic:
        raise NotSupportedManifestType(manifest, manifest_name=manifest.name, supported_type="dynamic", given_type="static")

    dataset_name = _parse_and_validate_dataset_name(context, manifest, params)
    tmp_path = await _create_and_validate_tmp_file(context, manifest, request)

    try:
        manifest_path = ManifestPath(type='csv', path=tmp_path)
        target_context = _setup_context(context, manifest_path)
        store = prepare_manifest(target_context, ensure_config_dir=True, full_load=True)
        target_manifest = store.manifest
        validate_manifest(context, target_manifest, dataset_name)

        rename_data = create_migrate_rename_mapping(context, target_context, manifest, target_manifest, dataset_name)
        migrate_meta = MigrateMeta(
            plan=False,
            autocommit=False,
            rename=MigrateRename(
                rename_src=rename_data
            ),
            datasets=[dataset_name],
            migration_extension=(lambda: commands.update_manifest_dataset_schema(context, manifest, target_manifest))
        )
        commands.migrate(context, target_manifest, migrate_meta)
        backend = manifest.backend
        commands.reload_backend_metadata(context, manifest, backend)
        reset_affected_objects(context, manifest, dataset_name)

        _clean_up_file(tmp_path)
        return JSONResponse({"status": "ok"}, status_code=200)
    except Exception as e:
        _clean_up_file(tmp_path)
        raise e



from itertools import islice
from itertools import chain
from typing import Iterable, Any
from typing import Iterator
from typing import List
from typing import TypeVar
from typing import Union


def consume(generator, n=None):
    generator = islice(generator, n) if n else generator
    return sum((1 for _ in generator), 0)


def drain(g):
    for x in g:
        pass


def chunks(it, n=100):
    yield from iter(lambda: list(islice(it, n)), [])


def schunks(it, s=100):
    blen = 0
    buff = []
    for v in it:
        vlen = len(v)
        if blen + vlen > s and buff:
            yield buff
            blen = 0
            buff = []
        buff.append(v)
        blen += vlen
    if buff:
        yield buff


def peek(it):
    peek = list(islice(it, 1))
    return chain(peek, it)


def recursive_keys(dct, dot_notation=False, prefix=None):
    # yields all keys from a given nested dictionaries
    # e.g.:
    # d = dict(a=1, b=dict(c=2))
    # list(recursive_keys(dct, dot_notation=True))
    # >>> ['a', 'b.c']
    for k, v in dct.items():
        if isinstance(v, dict):
            if dot_notation:
                if prefix:
                    k = f"{prefix}{k}"
                prefix = f"{k}."
            yield from recursive_keys(v, dot_notation=dot_notation, prefix=prefix)
            prefix = None
        elif isinstance(v, list):
            # if there are lists - take first value from the list
            # assume that schema for all list elements are the same
            if dot_notation:
                if prefix:
                    k = f"{prefix}{k}"
                prefix = f"{k}."
            if v and isinstance(v[0], dict):
                yield from recursive_keys(v[0], dot_notation=dot_notation, prefix=prefix)
            else:
                yield k
            prefix = None
        else:  # do not yield dict's root key as it's redundant
            if prefix:
                k = f"{prefix}{k}"
            yield k


def last(it, default=None):
    res = default
    for res in it:
        pass
    return res


T = TypeVar('T')


def flatten(it: Iterable[Union[T, List[T]]]) -> Iterator[T]:
    for x in it:
        if isinstance(x, list):
            yield from x
        else:
            yield x


# Based on https://stackoverflow.com/a/54500174
# This is one of the fastest ways to get first value
def first_dict_value(it: dict, default=None):
    res = default
    for res in it:
        return it[res]
    return res


# Based on https://stackoverflow.com/a/54500174
# This is one of the fastest ways to get first key
def first_dict_key(it: dict, default=None):
    res = default
    for res in it:
        return res
    return res


def ensure_list(value: Any) -> List[Any]:
    if isinstance(value, tuple):
        return list(value)
    if isinstance(value, list):
        return value
    else:
        return [value]




import datetime
import itertools
import json
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Tuple

import sqlalchemy as sa

from spinta import spyna
from spinta.cli.helpers.push import prepare_data_for_push_state
from spinta.cli.helpers.push.components import PushRow, Saved
from spinta.cli.helpers.push.utils import get_data_checksum
from spinta.components import Context, pagination_enabled
from spinta.components import Model
from spinta.utils.json import fix_data_for_json
from spinta.utils.sqlite import migrate_table


def init_push_state(
    dburi: str,
    models: List[Model],
) -> Tuple[sa.engine.Engine, sa.MetaData]:
    engine = sa.create_engine(dburi)
    metadata = sa.MetaData(engine)
    inspector = sa.inspect(engine)

    page_table = sa.Table(
        '_page', metadata,
        sa.Column('model', sa.Text, primary_key=True),
        sa.Column('property', sa.Text),
        sa.Column('value', sa.Text),
    )
    page_table.create(checkfirst=True)

    types = {
        'string': sa.Text,
        'date': sa.Date,
        'datetime': sa.DateTime,
        'time': sa.Time,
        'integer': sa.Integer,
        'number': sa.Numeric
    }

    for model in models:
        pagination_cols = []
        if pagination_enabled(model):
            for prop in model.page.keys.values():
                _type = types.get(prop.dtype.name, sa.Text)
                pagination_cols.append(
                    sa.Column(f"page.{prop.name}", _type, index=True)
                )

        table = sa.Table(
            model.name, metadata,
            sa.Column('id', sa.Unicode, primary_key=True),
            sa.Column('checksum', sa.Unicode),
            sa.Column('revision', sa.Unicode),
            sa.Column('pushed', sa.DateTime),
            sa.Column('error', sa.Boolean),
            sa.Column('data', sa.Text),
            *pagination_cols
        )
        migrate_table(engine, metadata, inspector, table, renames={
            'rev': 'checksum',
        })

    return engine, metadata


def reset_pushed(
    context: Context,
    models: List[Model],
    metadata: sa.MetaData,
):
    conn = context.get('push.state.conn')
    for model in models:
        table = metadata.tables[model.name]

        # reset pushed so we could see which objects were deleted
        conn.execute(
            table.update().
            values(pushed=None)
        )


def check_push_state(
    context: Context,
    rows: Iterable[PushRow],
    metadata: sa.MetaData
):
    conn = context.get('push.state.conn')

    for model_type, group in itertools.groupby(rows, key=_get_model_type):
        saved_rows = {}
        if model_type:
            table = metadata.tables[model_type]

            query = sa.select([table.c.id, table.c.revision, table.c.checksum])
            saved_rows = {
                state[table.c.id]: Saved(
                    state[table.c.revision],
                    state[table.c.checksum],
                )
                for state in conn.execute(query)
            }

        for row in group:
            if row.send and not row.error and row.op != "delete":
                _id = row.data['_id']
                row.checksum = get_data_checksum(row.data, row.model)
                saved = saved_rows.get(_id)
                if saved is None:
                    row.op = "insert"
                    row.saved = False
                else:
                    row.op = "patch"
                    row.saved = True
                    row.data['_revision'] = saved.revision
                    if saved.checksum == row.checksum:
                        conn.execute(
                            table.update().
                            where(table.c.id == _id).
                            values(
                                pushed=datetime.datetime.now()
                            )
                        )
                        # Skip if no changes, but only if not paginated, since pagination already skips those
                        if '_page' not in row.data:
                            continue

            yield row


def save_push_state(
    context: Context,
    rows: Iterable[PushRow],
    metadata: sa.MetaData,
) -> Iterator[PushRow]:
    conn = context.get('push.state.conn')
    page_table = metadata.tables['_page']
    model_pagination_check = {}
    for row in rows:
        table = metadata.tables[row.data['_type']]
        model_name = row.model.model_type()
        if model_name not in model_pagination_check:
            model_pagination_check[model_name] = pagination_enabled(row.model)

        if model_pagination_check[model_name] and '_page' in row.data:
            loaded = row.data['_page']
            page = {
                prop.name: loaded[i]
                for i, prop in enumerate(row.model.page.keys.values())
            }
            page = prepare_data_for_push_state(context, row.model, page)
            save_page_values(
                conn=conn,
                table=page_table,
                model=row.model,
                page_data=page
            )
            page = {f'page.{key}': value for key, value in page.items()}
            row.data.pop('_page')
        else:
            page = {}

        if row.error and row.op != "delete":
            data = fix_data_for_json(row.data)
            data = json.dumps(data)
        else:
            data = None

        if '_id' in row.data:
            _id = row.data['_id']
        else:
            _id = spyna.parse(row.data['_where'])['args'][1]

        if row.op == "delete" and not row.error:
            conn.execute(
                table.delete().
                where(table.c.id == _id)
            )
        elif row.saved:
            conn.execute(
                table.update().
                where(table.c.id == _id).
                values(
                    id=_id,
                    revision=row.data.get('_revision'),
                    checksum=row.checksum,
                    pushed=datetime.datetime.now(),
                    error=row.error,
                    data=data,
                    **page
                )
            )
        else:
            conn.execute(
                table.insert().
                values(
                    id=_id,
                    revision=row.data.get('_revision'),
                    checksum=row.checksum,
                    pushed=datetime.datetime.now(),
                    error=row.error,
                    data=data,
                    **page
                )
            )
        yield row


def save_page_values(
    conn: sa.engine.Connection,
    table: sa.Table,
    model: Model,
    page_data: dict
):
    model_name = model.model_type()
    page_row = conn.execute(
        sa.select(table.c.model)
        .where(table.c.model == model_name)
    ).scalar()

    exists = page_row is not None
    page_values = {key: value for key, value in page_data.items() if value is not None}

    if not page_values:
        return

    page_values = fix_data_for_json(page_values)
    page_values = json.dumps(page_values)

    if exists:
        conn.execute(
            table.update().
            where(
                (table.c.model == model_name)
            ).
            values(
                value=page_values
            )
        )
    else:
        conn.execute(
            table.insert().
            values(
                model=model.name,
                property=','.join([prop.name for prop in model.page.keys.values()]),
                value=page_values
            )
        )


def _get_model_type(row: PushRow) -> str:
    return row.data['_type']



import dataclasses
import datetime
from collections.abc import Generator

from typer import echo
from typing import List, Iterable

from spinta import commands
from spinta.auth import authorized
from spinta.cli.helpers.errors import ErrorCounter
from spinta.components import Context, Model, Action, Property, Config
from spinta.datasets.backends.helpers import extract_values_from_row
from spinta.datasets.keymaps.components import KeyMap
import tqdm

from spinta.utils.response import get_request


def _build_changelog_url(
    server: str,
    model: str,
    offset_cid: int,
    limit: int
) -> str:
    url = f'{server}/{model}/:changes'
    if offset_cid:
        url = f'{url}/{offset_cid + 1}'
    url = f'{url}?limit({limit})'
    return url


def _fetch_changelog_data(
    config: Config,
    model: Model,
    client,
    server: str,
    offset_cid: int,
    error_counter: ErrorCounter,
    timeout: tuple[float, float]
):
    limit = config.sync_page_size
    offset = offset_cid
    while True:
        url = _build_changelog_url(
            server=server,
            model=model.model_type(),
            offset_cid=offset,
            limit=limit
        )
        status_code, resp = get_request(
            client,
            url,
            error_counter=error_counter,
            timeout=timeout
        )
        if status_code != 200:
            break

        data = resp.get('_data')
        if not data:
            break

        cid = offset
        i = 0
        for i, row in enumerate(data, start=1):
            cid = row['_cid']
            yield row

        # Do not iterate further if limit was not reached
        if i < limit:
            break

        offset = cid


def _cid_offset(
    keymap: KeyMap,
    model_keymaps: list,
    reset_cid: bool
) -> int:
    sync_cid = 0
    if not reset_cid:
        initial = True
        for keymap_name in model_keymaps:
            cid = keymap.get_last_synced_id(keymap_name)
            if not cid:
                sync_cid = 0
                break

            if initial:
                sync_cid = cid
                initial = False
            else:
                sync_cid = min(sync_cid, cid)
    return sync_cid


@dataclasses.dataclass
class KeymapData:
    key: str
    value: object
    identifier: str
    data: dict


def process_keymap_data(
    keymap: KeyMap,
    model: Model,
    data: Iterable,
    primary_keys: list[Property],
    counters: dict,
    dry_run: bool
) -> Generator[KeymapData]:
    for row in data:
        cid = row['_cid']
        if row['_op'] == 'insert':
            yield from (
                KeymapData(key=key, value=value, identifier=identifier, data=row)
                for key, value, identifier in
                sync_model_insert(
                    keymap=keymap,
                    model=model,
                    row=row,
                    primary_keys=primary_keys,
                    counters=counters,
                    dry_run=dry_run
                )
            )
        elif row['_op'] == 'patch' or row['_op'] == 'update':
            yield from (
                KeymapData(key=key, value=value, identifier=identifier, data=row)
                for key, value, identifier in
                sync_model_update(
                    keymap=keymap,
                    model=model,
                    row=row,
                    counters=counters,
                    dry_run=dry_run
                )
            )


def sync_model_insert(keymap: KeyMap, model: Model, row: dict, primary_keys: List[Property], counters: dict,
                      dry_run: bool):
    values = _extract_row_data_from_keys(row, primary_keys)
    value = list(values.values())
    if len(value) == 1:
        value = value[0]

    if not dry_run:
        yield model.model_type(), value, row['_id']

    if model.model_type() in counters and model.model_type() in counters[model.model_type()]:
        counters[model.model_type()][model.model_type()].update(1)
    if '_total' in counters:
        counters['_total'].update(1)

    if row['_id'] and model.required_keymap_properties:
        for combination in model.required_keymap_properties:
            joined = '_'.join(combination)
            key = f'{model.model_type()}.{joined}'
            val = extract_values_from_row(row, model, combination)

            if not dry_run:
                yield key, val, row['_id']

            if model.model_type() in counters and key in counters[model.model_type()]:
                counters[model.model_type()][key].update(1)
            if '_total' in counters:
                counters['_total'].update(1)


def sync_model_update(keymap: KeyMap, model: Model, row: dict, counters: dict, dry_run: bool):
    for km, props in keymaps_affected_by_change(row, model).items():
        decoded = keymap.decode(km, row['_id'])
        remapped = remap_decoded_values(decoded, props)
        for key in remapped.keys():
            if key in row:
                remapped[key] = row[key]
        val = list(remapped.values())
        if len(val) == 1:
            val = val[0]

        if not dry_run:
            yield km, val, row['_id']

        if model.model_type() in counters and km in counters[model.model_type()]:
            counters[model.model_type()][km].update(1)
        if '_total' in counters:
            counters['_total'].update(1)


def sync_keymap(
    context: Context,
    keymap: KeyMap,
    client,
    server: str,
    models: List[Model],
    error_counter: ErrorCounter,
    no_progress_bar: bool,
    reset_cid: bool,
    timeout: tuple[float, float],
    dry_run: bool = False,
):
    config = context.get('config')
    counters = {}
    if not no_progress_bar:
        counters = {
            '_total': tqdm.tqdm(desc='SYNCHRONIZING KEYMAP', ascii=True)
        }
    try:
        for model in models:
            model_keymaps = [model.model_type()]
            primary_keys = model.external.pkeys
            skip_model = False
            for key in primary_keys:
                is_authorized = authorized(context, key, action=Action.SEARCH)
                if not is_authorized:
                    echo(f"SKIPPED KEYMAP '{model.model_type()}' MODEL SYNC, NO PERMISSION.")
                    skip_model = True
                    break
            if skip_model:
                continue

            for combination in model.required_keymap_properties:
                model_keymaps.append(f'{model.model_type()}.{"_".join(combination)}')

            # Get min cid (in case new model was added, or models are out of sync)
            offset_cid = _cid_offset(
                keymap=keymap,
                model_keymaps=model_keymaps,
                reset_cid=reset_cid
            )

            data = _fetch_changelog_data(
                config=config,
                model=model,
                client=client,
                server=server,
                offset_cid=offset_cid,
                error_counter=error_counter,
                timeout=timeout
            )

            if not no_progress_bar:
                model_counters = {}
                for keymap_name in model_keymaps:
                    model_counters[keymap_name] = tqdm.tqdm(desc=keymap_name, ascii=True)
                counters[model.model_type()] = model_counters

            data = process_keymap_data(
                keymap=keymap,
                model=model,
                data=data,
                primary_keys=primary_keys,
                counters=counters,
                dry_run=dry_run
            )
            data = commands.sync(context, keymap, data=data)
            try:
                for row in data:
                    offset_cid = row.data['_cid']
            finally:
                if not dry_run:
                    for keymap_name in model_keymaps:
                        keymap.update_sync_data(keymap_name, offset_cid, datetime.datetime.now())

                if model.model_type() in counters:
                    for counter in counters[model.model_type()].values():
                        counter.close()
    finally:
        if '_total' in counters:
            counters['_total'].close()


def remap_decoded_values(decoded: object, properties: List[str]):
    mapped = {}
    if not isinstance(decoded, list):
        decoded = [decoded]
    for i, prop in enumerate(properties):
        mapped[prop] = decoded[i]
    return mapped


def keymaps_affected_by_change(row: dict, model: Model) -> dict:
    keymaps = {}
    for key in model.external.pkeys:
        if key.name in row.keys():
            keymaps[model.model_type()] = list([prop.name for prop in model.external.pkeys])
            break

    for combination in model.required_keymap_properties:
        for prop in combination:
            if prop in row.keys():
                keymaps[f'{model.model_type()}.{"_".join(combination)}'] = combination
                break

    return keymaps


def _extract_row_data_from_keys(row: dict, keys: List[Property]) -> dict:
    result = {}
    for key in keys:
        if key.name in row:
            result[key.name] = row[key.name]
    return result


@commands.sync.register(Context, KeyMap)
def sync(context: Context, keymap: KeyMap, *, data: Generator[KeymapData]):
    for row in data:
        keymap.synchronize(row.key, row.value, row.identifier)
        yield row



from typing import List, Any, Tuple, Dict

from spinta.components import Page, Property
from spinta.core.ufuncs import ufunc, Expr, Negative, Bind, GetAttr
from spinta.datasets.backends.sql.ufuncs.components import Selected
from spinta.datasets.components import ExternalBackend
from spinta.exceptions import InvalidArgumentInExpression, CannotSelectTextAndSpecifiedLang, \
    LangNotDeclared, FieldNotInResource
from spinta.types.datatype import DataType, String, Ref, Object, Array, File, BackRef, PrimaryKey, ExternalRef
from spinta.types.text.components import Text
from spinta.ufuncs.basequerybuilder.components import BaseQueryBuilder, Star, ReservedProperty, NestedProperty, \
    ResultProperty, LiteralProperty, Flip
from spinta.ufuncs.basequerybuilder.helpers import get_pagination_compare_query, process_literal_value
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.schema import NA

# This file contains reusable resolvers, that should be backend independent
# in case there are cases where you need to have backend specific, just overload them
# keep in mind, that `select(env, expr)` is written backend specific, at least for now

# The main goal of this resolver logic, is that you have `_resolve_getattr`
# which is responsible for returning right types to process further
# Bind mean that it's a leaf node and can be accessed directly
# GetAttr mean that it can be nested further
# ForeignProperty is used, when you need to access properties from other models
# ReservedProperty is used, when datatypes do not have direct association to required field
#   example: File (when accessing child properties), Ref (when accessing it's own key, without needing to join models)


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


@ufunc.resolver(BaseQueryBuilder, Bind, Bind, name='getattr')
def getattr_(
    env: BaseQueryBuilder,
    field: Bind,
    attr: Bind
):
    return GetAttr(field.name, attr)


@ufunc.resolver(BaseQueryBuilder, Bind, GetAttr, name='getattr')
def getattr_(
    env: BaseQueryBuilder,
    obj: Bind,
    attr: GetAttr
):
    return GetAttr(obj.name, attr)


@ufunc.resolver(BaseQueryBuilder, GetAttr, Bind, name='getattr')
def getattr_(
    env: BaseQueryBuilder,
    obj: GetAttr,
    attr: Bind
):
    leaf = env.call('getattr', obj.name, attr)
    return GetAttr(obj.obj, leaf)


@ufunc.resolver(BaseQueryBuilder, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    attr: GetAttr,
) -> ForeignProperty:
    prop = env.model.properties[attr.obj]
    return env.call('_resolve_getattr', prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, Ref, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, Ref, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(BaseQueryBuilder, Ref, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Ref,
    attr: GetAttr,
):
    if attr.obj in dtype.properties:
        return dtype.properties[attr.obj].dtype
    prop = dtype.model.properties[attr.obj]
    fpr = ForeignProperty(None, dtype, prop.dtype)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, Ref, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Ref,
    attr: Bind,
):
    if attr.name in dtype.properties:
        return dtype.properties[attr.name].dtype

    # Check for self reference, no need to do joins if table already contains the value
    if attr.name == '_id':
        return ReservedProperty(dtype, attr.name)

    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(BaseQueryBuilder, ExternalRef, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: ExternalRef,
    attr: Bind,
):
    if attr.name in dtype.properties:
        return dtype.properties[attr.name].dtype

    # Check for self reference, no need to do joins if table already contains the value
    for refprop in dtype.refprops:
        if refprop.name == attr.name:
            return ReservedProperty(dtype, attr.name)

    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(BaseQueryBuilder, BackRef, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: BackRef,
    attr: GetAttr,
):
    prop = dtype.model.properties[attr.obj]
    fpr = ForeignProperty(None, dtype, prop.dtype)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, BackRef, Bind)
def _resolve_getattr(env, dtype, attr):
    prop = dtype.model.properties[attr.name]
    return ForeignProperty(None, dtype, prop.dtype)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, BackRef, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: BackRef,
    attr: GetAttr,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.obj]
    fpr = fpr.push(prop)
    return env.call('_resolve_getattr', fpr, prop.dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, BackRef, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: BackRef,
    attr: Bind,
) -> ForeignProperty:
    prop = dtype.model.properties[attr.name]
    return fpr.push(prop)


@ufunc.resolver(BaseQueryBuilder, Text, Bind)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Text,
    bind: Bind
):
    if dtype.prop.name in env.model.properties:
        prop = env.model.properties[dtype.prop.name]
        if bind.name in prop.dtype.langs:
            return prop.dtype.langs[bind.name].dtype
        raise LangNotDeclared(dtype, lang=bind.name)


@ufunc.resolver(BaseQueryBuilder, Object, GetAttr)
def _resolve_getattr(
    env: BaseQueryBuilder,
    dtype: Object,
    attr: GetAttr,
):
    if attr.obj in dtype.properties:
        prop = dtype.properties[attr.obj]

        return NestedProperty(
            left=dtype,
            right=env.call('_resolve_getattr', prop.dtype, attr.name)
        )
    raise FieldNotInResource(dtype, property=attr.obj)


@ufunc.resolver(BaseQueryBuilder, Object, Bind)
def _resolve_getattr(env, dtype, attr):
    if attr.name in dtype.properties:
        return NestedProperty(
            left=dtype,
            right=dtype.properties[attr.name].dtype
        )
    else:
        raise FieldNotInResource(dtype, property=attr.name)


@ufunc.resolver(BaseQueryBuilder, Array, (Bind, GetAttr))
def _resolve_getattr(env, dtype, attr):
    return NestedProperty(
        left=dtype,
        right=env.call('_resolve_getattr', dtype.items.dtype, attr)
    )


@ufunc.resolver(BaseQueryBuilder, File, Bind)
def _resolve_getattr(env, dtype, attr):
    return ReservedProperty(dtype, attr.name)


@ufunc.resolver(BaseQueryBuilder, GetAttr)
def select(env: BaseQueryBuilder, attr: GetAttr) -> Selected:
    resolved = env.call('_resolve_getattr', attr)
    return env.call('select', resolved)


@ufunc.resolver(BaseQueryBuilder, NestedProperty)
def select(env: BaseQueryBuilder, nested: NestedProperty) -> Selected:
    return Selected(
        prop=nested.left.prop,
        prep=env.call('select', nested.right),
    )


@ufunc.resolver(BaseQueryBuilder, ReservedProperty)
def select(env, prop):
    return env.call('select', prop.dtype, prop.param)


@ufunc.resolver(BaseQueryBuilder, ResultProperty)
def select(env: BaseQueryBuilder, prop: ResultProperty):
    return Selected(
        prep=prop.expr
    )


@ufunc.resolver(BaseQueryBuilder, ForeignProperty)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
    return env.call('select', fpr, fpr.right.prop)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, Property)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
) -> Selected:
    resolved_key = fpr / prop
    if resolved_key not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external and prop.external.prepare is not NA:
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            result = env.call('select', fpr, prop.dtype, result)
        else:
            result = env.call('select', fpr, prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[resolved_key] = result
    return env.resolved[resolved_key]


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, DataType, object)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    prep: Any,
) -> Selected:
    result = env.call('select', fpr, prep)
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, tuple)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', fpr, v) for v in prep)


@ufunc.resolver(BaseQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: BaseQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', fpr, prop)
    else:
        result = [
            env.call('select', fpr, prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(BaseQueryBuilder, list)
def select(
    env: BaseQueryBuilder,
    prep: List[Any],
) -> List[Any]:
    return [env.call('select', v) for v in prep]


@ufunc.resolver(BaseQueryBuilder, tuple)
def select(
    env: BaseQueryBuilder,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', v) for v in prep)


@ufunc.resolver(BaseQueryBuilder, dict)
def select(
    env: BaseQueryBuilder,
    prep: Dict[str, Any],
) -> Dict[str, Any]:
    # TODO: Add tests.
    return {k: env.call('select', v) for k, v in prep.items()}


@ufunc.resolver(BaseQueryBuilder, LiteralProperty)
def select(
    env: BaseQueryBuilder,
    prep: LiteralProperty,
) -> Selected:
    return Selected(prep=prep.value)


@ufunc.resolver(BaseQueryBuilder, DataType, object)
def select(env: BaseQueryBuilder, dtype: DataType, prep: Any) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if not isinstance(env.backend, ExternalBackend) or (prop.external and prop.external.name):
            sel = env.call('select', dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        result = env.call('select', prep)
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(BaseQueryBuilder, Flip)
def select(env: BaseQueryBuilder, flip_: Flip):
    return env.call('select', flip_.dtype, flip_)


@ufunc.resolver(BaseQueryBuilder, GetAttr)
def sort(env, field):
    dtype = env.call('_resolve_getattr', field)
    return env.call('sort', dtype)


@ufunc.resolver(BaseQueryBuilder, Expr, name='paginate')
def paginate(env, expr):
    if len(expr.args) != 1:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')
    page = expr.args[0]
    if isinstance(page, Page):
        if page.enabled:
            if not page.filter_only:
                env.page.select = env.call('select', page)
                for by, page_by in page.by.items():
                    sorted_ = env.call('sort',
                                       Negative(page_by.prop.name) if by.startswith("-") else Bind(page_by.prop.name))
                    if sorted_ is not None:
                        if isinstance(sorted_, (list, set, tuple)):
                            env.page.sort += sorted_
                        else:
                            env.page.sort.append(sorted_)
            env.page.page_ = page
            env.page.size = page.size
            return env.resolve(get_pagination_compare_query(page))
    else:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')


@ufunc.resolver(BaseQueryBuilder, Expr, name='expand')
def expand(env, expr):
    result = []
    if expr.args:
        for arg in expr.args:
            resolved = env.resolve(arg)
            result.append(resolved)
        return result
    return None


@ufunc.resolver(BaseQueryBuilder, str, name='op')
def op_(env, arg: str):
    if arg == '*':
        return Star()
    else:
        raise NotImplementedError


@ufunc.resolver(BaseQueryBuilder, Expr, name='page')
def page_(env, expr):
    pass


@ufunc.resolver(BaseQueryBuilder, DataType, list)
def validate_dtype_for_select(env, dtype: DataType, selected_props: List[Property]):
    raise NotImplemented(f"validate_dtype_for_select with {dtype.name} is not implemented")


@ufunc.resolver(BaseQueryBuilder, Text, list)
def validate_dtype_for_select(env, dtype: Text, selected_props: List[Property]):
    for prop in selected_props:
        if dtype.prop.name == prop.name or prop.parent == dtype.prop:
            raise CannotSelectTextAndSpecifiedLang(dtype)


@ufunc.resolver(BaseQueryBuilder, String, list)
def validate_dtype_for_select(env, dtype: String, selected_props: List[Property]):
    if dtype.prop.parent and isinstance(dtype.prop.parent.dtype, Text):
        parent = dtype.prop.parent
        for prop in selected_props:
            if parent.name == prop.name or prop == dtype.prop:
                raise CannotSelectTextAndSpecifiedLang(parent.dtype)


@ufunc.resolver(BaseQueryBuilder, NestedProperty, object, names=COMPARE)
def compare(env: BaseQueryBuilder, op: str, nested: NestedProperty, value: object):
    return env.call(op, nested.right, value)


@ufunc.resolver(BaseQueryBuilder, Bind, object, names=COMPARE)
def compare(env, op, field, value):
    prop = env.model.get_from_flatprops(field.name)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(BaseQueryBuilder, Expr)
def testlist(env: BaseQueryBuilder, expr: Expr) -> tuple:
    args, kwargs = expr.resolve(env)
    result = []
    for arg in args:
        result.append(process_literal_value(arg))
    return tuple(result)


@ufunc.resolver(BaseQueryBuilder)
def flip(env: BaseQueryBuilder):
    return env.call('flip', env.this)


@ufunc.resolver(BaseQueryBuilder, Property)
def flip(env: BaseQueryBuilder, prop: Property):
    return env.call('flip', prop.dtype)


@ufunc.resolver(BaseQueryBuilder, DataType)
def flip(env: BaseQueryBuilder, dtype: DataType):
    return Expr('flip')



import logging

from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.xsd.helpers import read_schema
from spinta.manifests.xsd.components import XsdManifest

log = logging.getLogger(__name__)


@commands.load.register(Context, XsdManifest)
def load(
    context: Context,
    manifest: XsdManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
    full_load=False
):
    assert freezed, (
        "SqlManifest does not have unfreezed version of manifest."
    )

    if load_internal:
        target = into or manifest
        if not commands.has_model(context, target, '_schema'):
            store = context.get('store')
            commands.load(context, store.internal, into=target, full_load=full_load)

    dataset_name = context.get("rc").get("given_dataset_name")
    schemas = read_schema(context, manifest.path, manifest.prepare, dataset_name)

    if into:
        log.info(
            'Loading freezed manifest %r into %r from %s.',
            manifest.name,
            into.name,
            manifest.path,
        )
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        log.info(
            'Loading freezed manifest %r from %s.',
            manifest.name,
            manifest.path,
        )
        load_manifest_nodes(context, manifest, schemas)

    for source in manifest.sync:
        commands.load(
            context, source,
            into=into or manifest,
            freezed=freezed,
            rename_duplicates=rename_duplicates,
            load_internal=load_internal,
            full_load=full_load
        )



from spinta.core.ufuncs import ufunc
from spinta.ufuncs.basequerybuilder.components import Selected
from spinta.ufuncs.resultbuilder.components import ResultBuilder


@ufunc.resolver(ResultBuilder)
def count(env: ResultBuilder):
    pass


@ufunc.resolver(ResultBuilder, Selected, Selected)
def point(env: ResultBuilder, x: Selected, y: Selected) -> str:
    x = env.data[x.item]
    y = env.data[y.item]
    return f'POINT ({x} {y})'



import pathlib
from io import StringIO
from typing import Tuple
from typing import Union

from pytest import FixtureRequest

from spinta import commands
from spinta.cli.helpers.store import load_store
from spinta.cli.manifest import copy_manifest
from spinta.components import Context
from spinta.components import Store
from spinta.core.config import RawConfig
from spinta.core.config import configure_rc
from spinta.manifests.components import Manifest
from spinta.manifests.components import ManifestPath
from spinta.manifests.helpers import detect_manifest_from_path
from spinta.manifests.tabular.helpers import normalizes_columns
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.manifests.tabular.helpers import striptable
from spinta.testing.context import TestContext
from spinta.testing.context import create_test_context


def compare_manifest(manifest: Manifest, expected: str, context: Context = None) -> Tuple[str, str]:
    expected = striptable(expected)
    if not context:
        context = Context('empty')
    if expected:
        header = expected.splitlines()[0]
        cols = normalizes_columns(header.split('|'))
        sizes = {c: len(x) - 2 for c, x in zip(cols, f' {header} '.split('|'))}
        actual = render_tabular_manifest(context, manifest, cols, sizes=sizes)
    else:
        actual = render_tabular_manifest(context, manifest)
    return actual, expected


def _create_file_path_for_type(tmp_path: pathlib.Path, file_name: str, manifest_type: str):
    if manifest_type != 'internal_sql' and tmp_path is None:
        raise Exception(f"TMP_PATH IS REQUIRED FOR {manifest_type} MANIFEST")

    if manifest_type == 'internal_sql':
        return 'sqlite:///' + str(tmp_path / f'{file_name}.sqlite')
    elif manifest_type in ['csv', 'xml', 'xlsx', 'yaml']:
        return str(tmp_path / f'{file_name}.{manifest_type}')
    elif manifest_type in ['ascii', 'tabular']:
        return str(tmp_path / f'{file_name}.txt')
    else:
        raise Exception(f"NO SUPPORT FOR {manifest_type} MANIFEST")


def load_manifest_get_context(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str] = None,
    *,
    load_internal: bool = False,
    request: FixtureRequest = None,
    manifest_type: str = '',
    tmp_path: pathlib.Path = None,
    full_load: bool = True,
    ensure_config_dir: bool = False,
    wipe_data: bool = True,
    **kwargs,
) -> TestContext:
    temp_rc = configure_rc(rc, None, **kwargs)
    context = create_test_context(temp_rc, wipe_data=wipe_data)

    if isinstance(manifest, str) and '|' in manifest:
        if manifest_type and manifest_type != 'ascii':
            ascii_file = _create_file_path_for_type(tmp_path, '_temp_ascii_manifest', 'ascii')
            output_file = _create_file_path_for_type(tmp_path, '_temp_manifest', manifest_type)
            with open(ascii_file, 'w') as f:
                f.write(manifest)
            copy_manifest(
                context,
                manifests=[ascii_file],
                output=output_file,
                output_type=manifest_type
            )
            manifests = [output_file]
        else:
            manifests = [ManifestPath(
                type='ascii',
                file=StringIO(manifest),
            )]
    elif manifest:
        if isinstance(manifest, pathlib.Path):
            manifest = str(manifest)
        manifest_ = detect_manifest_from_path(rc, manifest)
        if manifest_type and manifest_.type != manifest_type:
            output_path = _create_file_path_for_type(tmp_path, '_temp_manifest', manifest_type)
            copy_manifest(
                context,
                manifests=[manifest],
                output=output_path,
                output_type=manifest_type
            )
            manifest = output_path
        manifests = [manifest]
    else:
        manifests = manifest
    manifest_type = manifest_type if manifest_type != 'internal_sql' else 'internal'
    rc = configure_rc(rc, manifests, manifest_type='inline' if manifest_type != 'internal' else manifest_type, **kwargs)
    context = create_test_context(rc, request, wipe_data=wipe_data)
    store = load_store(context, verbose=False, ensure_config_dir=ensure_config_dir)
    commands.load(context, store.manifest, load_internal=load_internal, full_load=full_load)
    commands.link(context, store.manifest)
    return context


def load_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    **kwargs,
) -> Manifest:
    context = load_manifest_get_context(rc, manifest, **kwargs)
    store: Store = context.get('store')
    return store.manifest


def load_manifest_and_context(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str] = None,
    **kwargs,
) -> Tuple[Context, Manifest]:
    context = load_manifest_get_context(rc, manifest, **kwargs)
    store: Store = context.get('store')
    return context, store.manifest


def prepare_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    *,
    request: FixtureRequest = None,
    load_internal: bool = True,
    **kwargs,
) -> Tuple[TestContext, Manifest]:
    context = load_manifest_get_context(
        rc, manifest,
        request=request,
        load_internal=load_internal,
        **kwargs,
    )
    store: Store = context.get('store')
    commands.check(context, store.manifest)
    commands.prepare(context, store.manifest)
    return context, store.manifest


def bootstrap_manifest(
    rc: RawConfig,
    manifest: Union[pathlib.Path, str],
    *,
    request: FixtureRequest = None,
    load_internal: bool = True,
    full_load: bool = True,
    wipe_data: bool = True,
    **kwargs,
) -> TestContext:
    context = load_manifest_get_context(
        rc, manifest,
        request=request,
        load_internal=load_internal,
        full_load=full_load,
        wipe_data=wipe_data,
        **kwargs,
    )
    store: Store = context.get('store')
    commands.check(context, store.manifest)
    commands.prepare(context, store.manifest)
    commands.bootstrap(context, store.manifest)
    context.loaded = True
    return context



from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.components import Context
from spinta.manifests.internal_sql.components import InternalSQLManifest


@commands.reload_backend_metadata.register(Context, InternalSQLManifest, PostgreSQL)
def reload_backend_metadata(context: Context, manifest: InternalSQLManifest, backend: PostgreSQL):
    backend.schema.clear()
    backend.tables = {}
    commands.prepare(context, backend, manifest, ignore_duplicate=True)



from typing import Dict, Any
from typing import Iterable
from typing import List

import requests
import sqlalchemy as sa
import tqdm

from spinta import commands
from spinta.cli.helpers.data import ModelRow
from spinta.cli.helpers.errors import ErrorCounter
from spinta.cli.helpers.push.utils import construct_where_condition_from_page
from spinta.cli.helpers.push.write import prepare_rows_with_errors
from spinta.components import Context, pagination_enabled
from spinta.components import Model
from spinta.components import get_page_size
from spinta.exceptions import InfiniteLoopWithPagination, TooShortPageSize


def get_rows_with_errors(
    client: requests.Session,
    server: str,
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    counts: Dict[str, int],
    retry: int,
    timeout: tuple[float, float],
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None
):
    rows = _iter_rows_with_errors(
        client,
        server,
        models,
        context,
        metadata,
        counts,
        timeout,
        no_progress_bar,
        error_counter,
    )
    if not no_progress_bar:
        rows = tqdm.tqdm(rows, f'RETRY #{retry}', ascii=True, total=sum(counts.values()))
    for row in rows:
        yield row


def get_rows_with_errors_counts(
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
) -> dict:
    counts = {}
    conn = context.get('push.state.conn')
    for model in models:
        table = metadata.tables[model.name]

        row_count = conn.execute(
            sa.select(sa.func.count(table.c.id)).
            where(
                table.c.error.is_(True)
            )
        )

        counts[model.name] = row_count.scalar()
    return counts


def _iter_rows_with_errors(
    client: requests.Session,
    server: str,
    models: List[Model],
    context: Context,
    metadata: sa.MetaData,
    counts: Dict[str, int],
    timeout: tuple[float, float],
    no_progress_bar: bool = False,
    error_counter: ErrorCounter = None,
) -> Iterable[ModelRow]:
    conn = context.get('push.state.conn')
    config = context.get('config')

    for model in models:
        size = get_page_size(config, model)
        table = metadata.tables[model.name]

        if pagination_enabled(model):
            rows = _get_error_rows_with_page(
                context,
                model,
                table,
                size,
            )
        else:
            rows = conn.execute(
                sa.select([table.c.id, table.c.checksum, table.c.data]).
                where(
                    table.c.error.is_(True)
                )
            )
        if not no_progress_bar:
            rows = tqdm.tqdm(
                rows,
                model.name,
                ascii=True,
                total=counts.get(model.name),
                leave=False
            )

        yield from prepare_rows_with_errors(
            client,
            server,
            context,
            rows,
            model,
            table,
            timeout,
            error_counter
        )


def _get_error_rows_with_page(
    context: Context,
    model: Model,
    table: sa.Table,
    size: int,
):
    conn = context.get('push.state.conn')
    order_by = []

    page = commands.create_page(model.page)
    page.size += 1

    for page_by in page.by.values():
        order_by.append(sa.asc(table.c[f"page.{page_by.prop.name}"]))

    required_where_cond = sa.and_(
        table.c.error.is_(True)
    )
    last_value = None
    while True:
        finished = True
        cond = construct_where_condition_from_page(page, table)

        where_cond = None
        if cond is not None:
            where_cond = cond

        if where_cond is not None:
            where_cond = sa.and_(
                required_where_cond,
                where_cond
            )
        else:
            where_cond = required_where_cond

        rows = conn.execute(
            sa.select([table]).
            where(
                where_cond
            ).order_by(*order_by).limit(size)
        )
        first_value = None
        previous_value = None
        for i, row in enumerate(rows):
            if i > size - 1:
                if row == previous_value:
                    raise TooShortPageSize(
                        page,
                        page_size=size,
                        page_values=previous_value
                    )
                break
            previous_value = row
            if finished:
                finished = False

            if first_value is None:
                first_value = row
                if first_value == last_value:
                    raise InfiniteLoopWithPagination(
                        page,
                        page_size=size,
                        page_values=first_value
                    )
                else:
                    last_value = first_value

            for by, page_by in page.by.items():
                page.update_value(by, page_by.prop, row[f"page.{page_by.prop.name}"])
            yield row

        if finished:
            break

